# -*- coding: utf-8 -*-
"""Interface for plist parser plugins.

Plist files are only one example of a type of object that the Plaso tool is
expected to encounter and process. There can be and are many other parsers which
are designed to process specific data types.

PlistPlugin defines the attributes necessary for registration, discovery and
operation of plugins for plist files which will be used by PlistParser.
"""

import abc
import plistlib
import uuid

import pytz

from dfdatetime import cocoa_time as dfdatetime_cocoa_time
from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.parsers import logger
from plaso.parsers import plugins


class NSKeyedArchiverDecoder(object):
  """Decoder for NSKeyedArchiver encoded plists.

  Also see:
    https://developer.apple.com/documentation/foundation/nskeyedarchiver
  """

  _CALLBACKS = {
      'NSArray': '_DecodeNSArray',
      'NSData': '_DecodeNSData',
      'NSDate': '_DecodeNSDate',
      'NSDictionary': '_DecodeNSDictionary',
      'NSHashTable': '_DecodeNSHashTable',
      'NSNull': '_DecodeNSNull',
      'NSObject': '_DecodeCompositeObject',
      'NSSet': '_DecodeNSArray',
      'NSString': '_DecodeNSString',
      'NSURL': '_DecodeNSURL',
      'NSUUID': '_DecodeNSUUID'}

  def _DecodeCompositeObject(
      self, plist_property, objects_array, parent_objects):
    """Decodes a composite object.

    Args:
      plist_property (object): property containing the encoded composite object.
      objects_array (list[object]): $objects array.
      parent_objects (list[int]): parent object UIDs.

    Returns:
      object: decoded composite object.

    Raises:
      RuntimeError: if the composite object cannot be decoded.
    """
    composite_object = {}

    for key in plist_property:
      if key == '$class':
        continue

      value_plist_property = plist_property.get(key, None)

      value_plist_uid = self._GetPlistUID(value_plist_property)
      if value_plist_uid is None:
        composite_object[key] = self._DecodeObject(
            value_plist_property, objects_array, parent_objects)
        continue

      if value_plist_uid in parent_objects:
        continue

      parent_objects.append(value_plist_uid)

      composite_object[key] = self._DecodeObject(
          objects_array[value_plist_uid], objects_array, parent_objects)

      parent_objects.pop(-1)

    return composite_object

  def _DecodeNSArray(self, plist_property, objects_array, parent_objects):
    """Decodes a NSArray or NSSet.

    Args:
      plist_property (object): property containing the encoded NSArray or NSSet.
      objects_array (list[object]): $objects array.
      parent_objects (list[int]): parent object UIDs.

    Returns:
      list[object]: decoded NSArray or NSSet.

    Raises:
      RuntimeError: if the NSArray or NSSet cannot be decoded.
    """
    class_name = self._GetClassName(plist_property, objects_array)
    if 'NS.objects' not in plist_property:
      raise RuntimeError(f'Missing NS.objects in {class_name:s}')

    ns_objects_property = plist_property['NS.objects']

    ns_array = []

    for index, ns_object_property in enumerate(ns_objects_property):
      ns_object_plist_uid = self._GetPlistUID(ns_object_property)
      if ns_object_plist_uid is None:
        raise RuntimeError(
            f'Missing UID in NS.objects[{index:d}] property of {class_name:s}.')

      if ns_object_plist_uid in parent_objects:
        continue

      ns_object_referenced_property = objects_array[ns_object_plist_uid]

      parent_objects.append(ns_object_plist_uid)

      ns_array_element = self._DecodeObject(
          ns_object_referenced_property, objects_array, parent_objects)

      parent_objects.pop(-1)

      ns_array.append(ns_array_element)

    return ns_array

  # pylint: disable=unused-argument

  def _DecodeNSData(self, plist_property, objects_array, parent_objects):
    """Decodes a NSData.

    Args:
      plist_property (object): property containing the encoded NSData.
      objects_array (list[object]): $objects array.
      parent_objects (list[int]): parent object UIDs.

    Returns:
      bytes: decoded NSData.

    Raises:
      RuntimeError: if the NSData cannot be decoded.
    """
    class_name = self._GetClassName(plist_property, objects_array)
    if 'NS.data' not in plist_property:
      raise RuntimeError(f'Missing NS.data in {class_name:s}')

    ns_data = plist_property['NS.data']

    if not isinstance(ns_data, bytes):
      type_string = type(ns_data)
      raise RuntimeError(
          f'Unsupported type: {type_string!s} in {class_name:s}.NS.data.')

    return ns_data

  def _DecodeNSDate(self, plist_property, objects_array, parent_objects):
    """Decodes a NSDate.

    Args:
      plist_property (object): property containing the encoded NSDate.
      objects_array (list[object]): $objects array.
      parent_objects (list[int]): parent object UIDs.

    Returns:
      dfdatetime.CocoaTime: decoded NSDate.

    Raises:
      RuntimeError: if the NSDate cannot be decoded.
    """
    class_name = self._GetClassName(plist_property, objects_array)
    if 'NS.time' not in plist_property:
      raise RuntimeError(f'Missing NS.time in {class_name:s}')

    ns_time = plist_property['NS.time']

    if not isinstance(ns_time, float):
      type_string = type(ns_time)
      raise RuntimeError(
          f'Unsupported type: {type_string!s} in {class_name:s}.NS.time.')

    return dfdatetime_cocoa_time.CocoaTime(timestamp=ns_time)

  # pylint: enable=unused-argument

  def _DecodeNSDictionary(self, plist_property, objects_array, parent_objects):
    """Decodes a NSDictionary.

    Args:
      plist_property (object): property containing the encoded NSDictionary.
      objects_array (list[object]): $objects array.
      parent_objects (list[int]): parent object UIDs.

    Returns:
      dict[str, object]: decoded NSDictionary.

    Raises:
      RuntimeError: if the NSDictionary cannot be decoded.
    """
    class_name = self._GetClassName(plist_property, objects_array)
    if 'NS.keys' not in plist_property or 'NS.objects' not in plist_property:
      raise RuntimeError(f'Missing NS.keys or NS.objects in {class_name:s}')

    ns_keys_property = plist_property['NS.keys']
    ns_objects_property = plist_property['NS.objects']

    if len(ns_keys_property) != len(ns_objects_property):
      raise RuntimeError((
          f'Mismatch between number of NS.keys and NS.objects in '
          f'{class_name:s}'))

    ns_dictionary = {}

    for index, ns_object_property in enumerate(ns_objects_property):
      ns_key_property = ns_keys_property[index]

      ns_key_plist_uid = self._GetPlistUID(ns_key_property)
      if ns_key_plist_uid is None:
        raise RuntimeError(
            f'Missing UID in NS.keys[{index:d}] property of {class_name:s}.')

      ns_key_referenced_property = objects_array[ns_key_plist_uid]

      parent_objects.append(ns_key_plist_uid)

      ns_key = self._DecodeObject(
          ns_key_referenced_property, objects_array, parent_objects)

      parent_objects.pop(-1)

      if not ns_key:
        raise RuntimeError((
            f'Missing {class_name:s}.NS.keys[{index:d}] with UID: '
            f'{ns_key_plist_uid:d}.'))

      if not isinstance(ns_key, str):
        type_string = type(ns_key)
        raise RuntimeError((
            f'Unsupported type: {type_string!s} in {class_name:s}.NS.keys'
            f'[{index:d}] with UID: {ns_key_plist_uid:d}.'))

      ns_object_plist_uid = self._GetPlistUID(ns_object_property)
      if ns_object_plist_uid is None:
        raise RuntimeError((
            f'Missing UID in NS.objects[{index:d}] property of {class_name:s}'
            f'.{ns_key:s}.'))

      if ns_object_plist_uid in parent_objects:
        continue

      ns_object_referenced_property = objects_array[ns_object_plist_uid]

      parent_objects.append(ns_object_plist_uid)

      ns_dictionary[ns_key] = self._DecodeObject(
          ns_object_referenced_property, objects_array, parent_objects)

      parent_objects.pop(-1)

    return ns_dictionary

  def _DecodeNSHashTable(self, plist_property, objects_array, parent_objects):
    """Decodes a NSHashTable.

    Args:
      plist_property (object): property containing the encoded NSHashTable.
      objects_array (list[object]): $objects array.
      parent_objects (list[int]): parent object UIDs.

    Returns:
      object: decoded NSHashTable.

    Raises:
      RuntimeError: if the NSHashTable cannot be decoded.
    """
    class_name = self._GetClassName(plist_property, objects_array)
    if '$1' not in plist_property:
      raise RuntimeError(f'Missing $1 in {class_name:s}')

    value_property = plist_property['$1']

    value_plist_uid = self._GetPlistUID(value_property)
    if value_plist_uid is None:
      type_string = type(value_property)
      raise RuntimeError(f'Unsupported {class_name:s}.$1 type: {type_string!s}')

    if value_plist_uid in parent_objects:
      raise RuntimeError((
          f'{class_name:s}.$1 wth UID: {value_plist_uid:d} in parent objects'))

    referenced_property = objects_array[value_plist_uid]
    if not referenced_property:
      raise RuntimeError(
          f'Missing {class_name:s}.$1 with UID: {value_plist_uid:d}.')

    # TODO: what about value $0? It seems to indicate the number of elements
    # in the hash table.
    # TODO: what about value $2?

    parent_objects.append(value_plist_uid)

    ns_hash_table = self._DecodeCompositeObject(
        referenced_property, objects_array, parent_objects)

    parent_objects.pop(-1)

    return ns_hash_table

  # pylint: disable=unused-argument

  def _DecodeNSNull(self, plist_property, objects_array, parent_objects):
    """Decodes a NSNull.

    Args:
      plist_property (object): property containing the encoded NSNull.
      objects_array (list[object]): $objects array.
      parent_objects (list[int]): parent object UIDs.

    Returns:
      None: decoded NSNull.
    """
    return None

  # pylint: enable=unused-argument

  def _DecodeNSObject(self, plist_property, objects_array, parent_objects):
    """Decodes a NSObject.

    Args:
      plist_property (object): property containing the encoded NSObject.
      objects_array (list[object]): $objects array.
      parent_objects (list[int]): parent object UIDs.

    Returns:
      object: decoded NSObject.

    Raises:
      RuntimeError: if the NSObject cannot be decoded.
    """
    class_property = plist_property.get('$class', None)
    if not class_property:
      raise RuntimeError('$class property missing in NSObject.')

    class_plist_uid = self._GetPlistUID(class_property)
    if class_plist_uid is None:
      raise RuntimeError('Missing UID in $class property of NSObject.')

    referenced_property = objects_array[class_plist_uid]
    if not referenced_property:
      raise RuntimeError(
          f'Missing NSObject.$class with UID: {class_plist_uid:d}.')

    class_name = referenced_property.get('$classname', None)
    if not class_name:
      raise RuntimeError((
          f'$classname property missing in NSObject.$class with UID: '
          f'{class_plist_uid:d}.'))

    classes = referenced_property.get('$classes', None)
    if not classes:
      raise RuntimeError((
          f'$classes property missing in NSObject.$class with UID: '
          f'{class_plist_uid:d}.'))

    for name in classes:
      callback_method = self._CALLBACKS.get(name, None)
      if callback_method:
        break

    if not callback_method:
      raise RuntimeError(f'Missing callback for class: {class_name:s}')

    callback = getattr(self, callback_method, None)
    return callback(plist_property, objects_array, parent_objects)

  # pylint: disable=unused-argument

  def _DecodeNSString(self, plist_property, objects_array, parent_objects):
    """Decodes a NSString.

    Args:
      plist_property (object): property containing the encoded NSString.
      objects_array (list[object]): $objects array.
      parent_objects (list[int]): parent object UIDs.

    Returns:
      str: decoded NSString.

    Raises:
      RuntimeError: if the NSString cannot be decoded.
    """
    class_name = self._GetClassName(plist_property, objects_array)
    if 'NS.string' not in plist_property:
      raise RuntimeError(f'Missing NS.string in {class_name:s}')

    ns_string = plist_property['NS.string']

    if not isinstance(ns_string, str):
      type_string = type(ns_string)
      raise RuntimeError(
          f'Unsupported type: {type_string!s} in {class_name:s}.NS.string.')

    return ns_string

  def _DecodeNSURL(self, plist_property, objects_array, parent_objects):
    """Decodes a NSURL.

    Args:
      plist_property (object): property containing the encoded NSURL.
      objects_array (list[object]): $objects array.
      parent_objects (list[int]): parent object UIDs.

    Returns:
      str: decoded NSURL.

    Raises:
      RuntimeError: if the NSURL cannot be decoded.
    """
    if 'NS.base' not in plist_property or 'NS.relative' not in plist_property:
      raise RuntimeError('Missing NS.base or NS.relative in NSURL')

    ns_base_plist_uid = self._GetPlistUID(plist_property['NS.base'])
    if ns_base_plist_uid is None:
      raise RuntimeError('Missing UID in NS.base property of NSURL.')

    ns_base = objects_array[ns_base_plist_uid]
    if not ns_base:
      raise RuntimeError(
          f'Missing NSURL.NS.base with UID: {ns_base_plist_uid:d}.')

    if not isinstance(ns_base, str):
      type_string = type(ns_base)
      raise RuntimeError((
          f'Unsupported type: {type_string!s} in NSURL.NS.base with UID: '
          f'{ns_base_plist_uid:d}.'))

    ns_relative_plist_uid = self._GetPlistUID(plist_property['NS.relative'])
    if ns_relative_plist_uid is None:
      raise RuntimeError('Missing UID in NS.relative property of NSURL.')

    ns_relative = objects_array[ns_relative_plist_uid]
    if not ns_relative:
      raise RuntimeError(
          f'Missing NSURL.NS.relative with UID: {ns_relative_plist_uid:d}.')

    if not isinstance(ns_relative, str):
      type_string = type(ns_relative)
      raise RuntimeError((
          f'Unsupported type: {type_string!s} in NSURL.NS.relative with UID: '
          f'{ns_relative_plist_uid:d}.'))

    if ns_base == '$null':
      ns_url = ns_relative
    else:
      ns_url = '/'.join([ns_base, ns_relative])

    return ns_url

  def _DecodeNSUUID(self, plist_property, objects_array, parent_objects):
    """Decodes a NSUUID.

    Args:
      plist_property (object): property containing the encoded NSUUID.
      objects_array (list[object]): $objects array.
      parent_objects (list[int]): parent object UIDs.

    Returns:
      str: decoded NSUUID.

    Raises:
      RuntimeError: if the NSUUID cannot be decoded.
    """
    if 'NS.uuidbytes' not in plist_property:
      raise RuntimeError('Missing NS.uuidbytes')

    ns_uuidbytes = plist_property['NS.uuidbytes']
    if len(ns_uuidbytes) != 16:
      raise RuntimeError('Unsupported NS.uuidbytes size')

    return str(uuid.UUID(bytes=ns_uuidbytes))

  # pylint: enable=unused-argument

  def _DecodeObject(self, plist_property, objects_array, parent_objects):
    """Decodes an object.

    Args:
      plist_property (object): property containing the encoded NSUUID.
      objects_array (list[object]): $objects array.
      parent_objects (list[int]): parent object UIDs.

    Returns:
      object: decoded object.

    Raises:
      RuntimeError: if the object cannot be decoded.
    """
    if isinstance(plist_property, dict) and '$class' in plist_property:
      return self._DecodeNSObject(plist_property, objects_array, parent_objects)

    if isinstance(plist_property, str) and plist_property == '$null':
      return None

    return plist_property

  def _GetClassName(self, plist_property, objects_array):
    """Retrieves a class name.

    Args:
      plist_property (object): property containing the $class property.
      objects_array (list[object]): $objects array.

    Returns:
      str: class name or None if not available.

    Raises:
      RuntimeError: if the class name cannot be retrieved.
    """
    class_property = plist_property.get('$class', None)
    if not class_property:
      raise RuntimeError('Missing $class property.')

    class_plist_uid = self._GetPlistUID(class_property)
    if class_plist_uid is None:
      raise RuntimeError('Missing UID in $class property.')

    referenced_property = objects_array[class_plist_uid]
    if not referenced_property:
      raise RuntimeError(f'Missing class with UID: {class_plist_uid:d}.')

    class_name = referenced_property.get('$classname', None)
    if not class_name:
      raise RuntimeError((
          f'$classname property missing in class with UID: '
          f'{class_plist_uid:d}.'))

    return class_name

  def _GetPlistUID(self, plist_property):
    """Retrieves a plist UID.

    Args:
      plist_property (object): property containing the plist UID.

    Returns:
      int: plist UID or None if not available.
    """
    if isinstance(plist_property, plistlib.UID):
      return plist_property.data

    if (isinstance(plist_property, dict) and 'CF$UID' in plist_property and
        len(plist_property) == 1):
      return plist_property['CF$UID']

    return None

  def Decode(self, root_item):
    """Decodes a NSKeyedArchiver encoded plist.

    Args:
      root_item (object): root object of the NSKeyedArchiver encoded plist.

    Returns:
      dict[str, object]: root object of the decoded plist.

    Raises:
      RuntimeError: if the plist cannot be decoded.
    """
    if not isinstance(root_item, dict):
      type_string = type(root_item)
      raise RuntimeError(
          f'Unsupported plist: unsupported root item type: {type_string!s}.')

    archiver = root_item.get('$archiver')
    version = root_item.get('$version')
    if archiver != 'NSKeyedArchiver' or version != 100000:
      raise RuntimeError(f'Unsupported plist: {archiver!s} {version!s}')

    decoded_object = {}

    objects_array = root_item.get('$objects') or []

    top_property = root_item.get('$top') or {}
    for name, value_property in top_property.items():
      value_plist_uid = self._GetPlistUID(value_property)
      if value_plist_uid is None:
        decoded_object[name] = value_property
        continue

      value_referenced_property = objects_array[value_plist_uid]
      if not value_referenced_property:
        raise RuntimeError(
            f'Missing $top["{name:s}"] with UID: {value_plist_uid:d}.')

      decoded_object[name] = self._DecodeObject(
          value_referenced_property, objects_array, [value_plist_uid])

    # The root $top appears to be internal only to the NSKeyedArchiver encoded
    # plist.
    if len(decoded_object) == 1:
      decoded_object = decoded_object.get('root', decoded_object)

    return decoded_object

  def IsEncoded(self, root_item):
    """Determines if a plist is NSKeyedArchiver encoded.

    Args:
      root_item (object): root object of the NSKeyedArchiver encoded plist.

    Returns:
      bool: True if NSKeyedArchiver encoded, False otherwise.
    """
    if not isinstance(root_item, dict):
      return False

    archiver = root_item.get('$archiver')
    version = root_item.get('$version')

    return archiver == 'NSKeyedArchiver' and version == 100000


class PlistPathFilter(object):
  """The plist path filter."""

  def __init__(self, filename):
    """Initializes a plist path filter.

    Args:
      filename (str): expected file name of the plist.
    """
    super(PlistPathFilter, self).__init__()
    self._filename_lower_case = filename.lower()

  def Match(self, filename_lower_case):
    """Determines if a plist filename matches the filter.

    Note that this method does a case insensitive comparison.

    Args:
      filename_lower_case (str): filename of the plist in lower case.

    Returns:
      bool: True if the filename matches the filter.
    """
    return bool(filename_lower_case == self._filename_lower_case)


class PrefixPlistPathFilter(PlistPathFilter):
  """The prefix plist path filter."""

  def Match(self, filename_lower_case):
    """Determines if a plist filename matches the filter.

    Note that this method does a case insensitive comparison.

    Args:
      filename_lower_case (str): filename of the plist in lower case.

    Returns:
      bool: True if the filename matches the filter.
    """
    return filename_lower_case.startswith(self._filename_lower_case)


class PlistPlugin(plugins.BasePlugin):
  """This is an abstract class from which plugins should be based.

  The following are the attributes and methods expected to be overridden by a
  plugin.

  Attributes:
    PLIST_PATH_FILTERS (set[PlistPathFilter]): plist path filters that should
         match for the plugin to process the plist.
    PLIST_KEY (set[str]): keys holding values that are necessary for processing.

  Please note, PLIST_KEY is case sensitive and for a plugin to match a
  plist file needs to contain at minimum the number of keys needed for
  processing.

  For example if a Plist file contains the following keys,
  {'foo': 1, 'bar': 2, 'opt': 3} with 'foo' and 'bar' being keys critical to
  processing define PLIST_KEY as ['foo', 'bar']. If 'opt' is only optionally
  defined it can still be accessed by manually processing self.top_level from
  the plugin.
  """

  NAME = 'plist_plugin'

  # This is expected to be overridden by the processing plugin, for example:
  # PLIST_PATH_FILTERS = frozenset(PlistPathFilter('com.apple.bluetooth.plist'))
  PLIST_PATH_FILTERS = frozenset()

  # PLIST_KEYS is a list of keys required by a plugin. This is expected to be
  # overridden by the processing plugin, for example:
  # PLIST_KEYS = frozenset(['DeviceCache', 'PairedDevices'])
  PLIST_KEYS = frozenset()

  def _GetDateTimeValueFromPlistKey(self, plist_key, plist_value_name):
    """Retrieves a date and time value from a specific value in a plist key.

    Args:
      plist_key (object): plist key.
      plist_value_name (str): name of the value in the plist key.

    Returns:
      dfdatetime.TimeElementsInMicroseconds: date and time or None if not
          available.
    """
    datetime_value = plist_key.get(plist_value_name, None)
    if not datetime_value:
      return None

    # dfDateTime relies on the time zone but since plistlib does not set one.
    datetime_value = datetime_value.replace(tzinfo=pytz.UTC)

    date_time = dfdatetime_time_elements.TimeElementsInMicroseconds()
    date_time.CopyFromDatetime(datetime_value)
    return date_time

  def _GetKeys(self, top_level, keys, depth=1):
    """Helper function to return keys nested in a plist dict.

    By default this function will return the values for the named keys requested
    by a plugin in match dictionary object. The default setting is to look
    a single level down from the root, which is suitable for most cases.

    For cases where there is variability in the name at the first level, for
    example the name is the MAC address of a device or a UUID, it is possible
    to override the depth limit and use GetKeys to fetch from a deeper level.

    E.g.
    Top_Level (root):                                             # depth = 0
    -- Key_Name_is_UUID_Generated_At_Install 1234-5678-8          # depth = 1
    ---- Interesting_SubKey_with_value_to_Process: [Values, ...]  # depth = 2

    Args:
      top_level (dict[str, object]): plist top-level key.
      keys (list[str]): names of the keys that should be returned.
      depth (int): number of levels to check for a match.

    Returns:
      dict[str, object]: the keys requested or an empty set if the plist is
          flat, for example the top level is a list instead of a dictionary.
    """
    match = {}
    if not isinstance(top_level, dict):
      # Return an empty dict here if top_level is a list object, which happens
      # if the plist file is flat.
      return match

    keys = set(keys)

    if depth == 1:
      for key in keys:
        match[key] = top_level.get(key, None)

    else:
      for _, parsed_key, parsed_value in self._RecurseKey(
          top_level, depth=depth):
        if parsed_key in keys:
          match[parsed_key] = parsed_value
          if set(match.keys()) == keys:
            return match

    return match

  def _RecurseKey(self, plist_item, depth=15, key_path=''):
    """Flattens nested dictionaries and lists by yielding its values.

    The hierarchy of a plist file is a series of nested dictionaries and lists.
    This is a helper function helps plugins navigate the structure without
    having to reimplement their own recursive methods.

    This method implements an overridable depth limit to prevent processing
    extremely deeply nested plists. If the limit is reached a debug message is
    logged indicating which key processing stopped on.

    Example Input Plist:
      plist_item = { DeviceRoot: { DeviceMAC1: [Value1, Value2, Value3],
                                   DeviceMAC2: [Value1, Value2, Value3]}}

    Example Output:
      ('', DeviceRoot, {DeviceMACs...})
      (DeviceRoot, DeviceMAC1, [Value1, Value2, Value3])
      (DeviceRoot, DeviceMAC2, [Value1, Value2, Value3])

    Args:
      plist_item (object): plist item to be checked for additional nested items.
      depth (Optional[int]): current recursion depth. This value is used to
          ensure we stop at the maximum recursion depth.
      key_path (Optional[str]): path of the current working key.

    Yields:
      tuple[str, str, object]: key path, key name and value.
    """
    if depth < 1:
      logger.debug(
          'Maximum recursion depth of 15 reached for key: {0:s}'.format(
              key_path))

    elif isinstance(plist_item, (list, tuple)):
      for sub_plist_item in plist_item:
        yield from self._RecurseKey(
            sub_plist_item, depth=depth - 1, key_path=key_path)

    elif hasattr(plist_item, 'items'):
      for subkey_name, value in plist_item.items():
        yield key_path, subkey_name, value

        if isinstance(value, dict):
          value = [value]
        elif not isinstance(value, (list, tuple)):
          continue

        for sub_plist_item in value:
          if isinstance(sub_plist_item, dict):
            subkey_path = '/'.join([key_path, subkey_name])
            yield from self._RecurseKey(
                sub_plist_item, depth=depth - 1, key_path=subkey_path)

  # pylint: disable=arguments-differ
  @abc.abstractmethod
  def _ParsePlist(
      self, parser_mediator, match=None, top_level=None, **unused_kwargs):
    """Extracts events from the values of entries within a plist.

    This is the main method that a plist plugin needs to implement.

    The contents of the plist keys defined in PLIST_KEYS will be made available
    to the plugin as self.matched{'KEY': 'value'}. The plugin should implement
    logic to parse this into a useful event for incorporation into the Plaso
    timeline.

    For example if you want to note the timestamps of when devices were
    LastInquiryUpdated you would need to examine the bluetooth config file
    called 'com.apple.bluetooth' and need to look at devices under the key
    'DeviceCache'. To do this the plugin needs to define:
        PLIST_PATH_FILTERS = frozenset([
            interface.PlistPathFilter('com.apple.bluetooth')])
        PLIST_KEYS = frozenset(['DeviceCache']).

    When a file with this key is encountered during processing self.matched is
    populated and the plugin's _ParsePlist() is called. The plugin would have
    self.matched = {'DeviceCache': [{'DE:AD:BE:EF:01': {'LastInquiryUpdate':
    DateTime_Object}, 'DE:AD:BE:EF:01': {'LastInquiryUpdate':
    DateTime_Object}'...}]} and needs to implement logic here to extract
    values, format, and produce the data as an event.PlistEvent.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      match (Optional[dict[str: object]]): keys extracted from PLIST_KEYS.
      top_level (Optional[dict[str, object]]): plist top-level item.
    """

  def CheckRequiredFormat(self, top_level):
    """Check if the plist has the minimal structure required by the plugin.

    Args:
      top_level (dict[str, object]): plist top-level item.

    Returns:
      bool: True if this is the correct plugin, False otherwise.
    """
    return set(top_level.keys()).issuperset(self.PLIST_KEYS)

  def Process(self, parser_mediator, top_level=None, **kwargs):
    """Extracts events from a plist file.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      top_level (Optional[dict[str, object]]): plist top-level item.
    """
    # This will raise if unhandled keyword arguments are passed.
    super(PlistPlugin, self).Process(parser_mediator)

    match = self._GetKeys(top_level, self.PLIST_KEYS)
    self._ParsePlist(parser_mediator, match=match, top_level=top_level)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for Tango on Android plugins."""

import unittest

from plaso.parsers.sqlite_plugins import android_tango

from tests.parsers.sqlite_plugins import test_lib


class AndroidTangoProfileTest(test_lib.SQLitePluginTestCase):
  """Tests for Tango on Android profile database plugin."""

  def testProcess(self):
    """Test the Process function on a Tango Android file."""
    plugin = android_tango.AndroidTangoProfilePlugin()
    storage_writer = self._ParseDatabaseFileWithPlugin(
        ['tango_android_profile.db'], plugin)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 57)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'asccess_time': None,
        'birthday': '1980-10-01',
        'data_type': 'android:tango:contact',
        'distance': 39.04880905,
        'first_name': 'Rouel',
        'friend_request_message': 'I am following you on Tango',
        'friend_request_time': '2016-01-15T14:35:20.436+00:00',
        'friend_request_type': 'outRequest',
        'gender': 'male',
        'is_friend': False,
        'last_active_time': '2016-01-15T13:21:45.624+00:00',
        'last_name': 'Henry',
        'status': 'Praying!'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 2)
    self.CheckEventData(event_data, expected_event_values)


class AndroidTangoTCTest(test_lib.SQLitePluginTestCase):
  """Tests for Tango on Android tc databases plugin."""

  def testProcess(self):
    """Test the Process function on a Tango Android file."""
    plugin = android_tango.AndroidTangoTCPlugin()
    storage_writer = self._ParseDatabaseFileWithPlugin(
        ['tango_android_tc.db'], plugin)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 25)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    # Test the a conversation entry.
    expected_event_values = {
        'conversation_identifier': 'DyGWr_010wQM_ozkIe-9Ww',
        'data_type': 'android:tango:conversation'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 2)
    self.CheckEventData(event_data, expected_event_values)

    # Test a message creation entry.
    expected_event_values = {
        'data_type': 'android:tango:message',
        'creation_time': '2016-01-15T14:41:33.027+00:00',
        'direction': 2,
        'message_identifier': 16777224,
        'sent_time': '2016-01-15T14:41:34.238+00:00'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 14)
    self.CheckEventData(event_data, expected_event_values)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the PE file parser."""

import os
import unittest

from dfvfs.lib import definitions as dfvfs_definitions
from dfvfs.path import factory as path_spec_factory
from dfvfs.resolver import resolver as path_spec_resolver

from plaso.containers import artifacts
from plaso.parsers import pe
from plaso.parsers import mediator as parsers_mediator

from tests.parsers import test_lib


class PECOFFTest(test_lib.ParserTestCase):
  """Tests for the PE file parser."""

  # pylint: disable=protected-access

  def testParseFileObjectOnExecutable(self):
    """Tests the ParseFileObject on a PE executable (EXE) file."""
    parser = pe.PEParser()
    storage_writer = self._ParseFile(['test_pe.exe'], parser)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 3)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'creation_time': '2015-04-21T14:53:56+00:00',
        'data_type': 'pe_coff:file',
        'export_table_modification_time': None,
        'imphash': '8d0739063fc8f9955cc6696b462544ab',
        'load_configuration_table_modification_time': None,
        'pe_type': 'Executable (EXE)'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 0)
    self.CheckEventData(event_data, expected_event_values)

    expected_event_values = {
        'data_type': 'pe_coff:dll_import',
        'delayed_import': False,
        'modification_time': '2015-04-21T14:53:55+00:00',
        'name': 'KERNEL32.dll'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 1)
    self.CheckEventData(event_data, expected_event_values)

    expected_event_values = {
        'data_type': 'pe_coff:dll_import',
        'delayed_import': True,
        'modification_time': '2015-04-21T14:53:54+00:00',
        'name': 'USER32.dll'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 2)
    self.CheckEventData(event_data, expected_event_values)

  def testParseFileObjectOnDriver(self):
    """Tests the ParseFileObject on a PE driver (SYS) file."""
    parser = pe.PEParser()
    storage_writer = self._ParseFile(['test_driver.sys'], parser)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 1)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'data_type': 'pe_coff:file',
        'creation_time': '2015-04-21T14:53:54+00:00',
        'export_table_modification_time': None,
        'imphash': 'd9c9c4541168665f44917e3ddc4a00d5',
        'load_configuration_table_modification_time': None,
        'pe_type': 'Driver (SYS)'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 0)
    self.CheckEventData(event_data, expected_event_values)

  def testParseFileObjectOnResourceFile(self):
    """Tests the ParseFileObject on a resource Dynamic Link Library file."""
    test_file_path = self._GetTestFilePath(['wrc-test-wevt_template.dll'])
    self._SkipIfPathNotExists(test_file_path)

    parser_mediator = parsers_mediator.ParserMediator()

    test_event_provider = artifacts.WindowsEventLogMessageFileArtifact()
    parser_mediator._extract_winevt_resources = True
    parser_mediator._windows_event_log_providers_per_path = {
        os.path.dirname(test_file_path).lower(): {
            'wrc-test-wevt_template.dll': test_event_provider}}

    storage_writer = self._CreateStorageWriter()
    parser_mediator.SetStorageWriter(storage_writer)

    path_spec = path_spec_factory.Factory.NewPathSpec(
        dfvfs_definitions.TYPE_INDICATOR_OS, location=test_file_path)
    file_entry = path_spec_resolver.Resolver.OpenFileEntry(path_spec)
    parser_mediator.SetFileEntry(file_entry)

    parser = pe.PEParser()

    file_object = file_entry.GetFileObject()
    parser.Parse(parser_mediator, file_object)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 1)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'creation_time': '2022-01-03T08:55:07+00:00',
        'data_type': 'pe_coff:file',
        'export_table_modification_time': None,
        'imphash': '1913ea9cbfeed7fd2a2ef823b6656f85',
        'load_configuration_table_modification_time': None,
        'pe_type': 'Dynamic Link Library (DLL)'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 0)
    self.CheckEventData(event_data, expected_event_values)

    number_of_artifacts = storage_writer.GetNumberOfAttributeContainers(
        'windows_eventlog_message_string')
    self.assertEqual(number_of_artifacts, 1)

    attribute_containers = list(storage_writer.GetAttributeContainers(
        'windows_eventlog_message_string'))
    self.assertEqual(attribute_containers[0].message_identifier, 0xb0010001)
    self.assertEqual(attribute_containers[0].language_identifier, 1033)
    self.assertEqual(attribute_containers[0].string, 'A test value')

    number_of_artifacts = storage_writer.GetNumberOfAttributeContainers(
        'windows_wevt_template_event')
    self.assertEqual(number_of_artifacts, 1)

    attribute_containers = list(storage_writer.GetAttributeContainers(
        'windows_wevt_template_event'))
    self.assertEqual(attribute_containers[0].identifier, 1)
    self.assertEqual(attribute_containers[0].message_identifier, 0xb0010001)
    self.assertEqual(
        attribute_containers[0].provider_identifier,
        '{67883bbc-d592-4d02-8e29-66907fcb07d6}')
    self.assertEqual(attribute_containers[0].version, 1)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the XLSX output module."""

import datetime
import os
import unittest
import zipfile

from defusedxml import ElementTree

from plaso.containers import events
from plaso.lib import definitions
from plaso.output import xlsx

from tests import test_lib as shared_test_lib
from tests.containers import test_lib as containers_test_lib
from tests.output import test_lib


class XLSXOutputModuleTest(test_lib.OutputModuleTestCase):
  """Test the XLSX output module."""

  # pylint: disable=protected-access

  _SHARED_STRINGS = 'xl/sharedStrings.xml'
  _SHEET1 = 'xl/worksheets/sheet1.xml'

  _COLUMN_TAG = '}c'
  _ROW_TAG = '}row'
  _SHARED_STRING_TAG = '}t'
  _SHARED_STRING_TYPE = 's'
  _TYPE_ATTRIBUTE = 't'
  _VALUE_STRING_TAG = '}v'

  _TEST_EVENTS = [
      {'data_type': 'test:event',
       'hostname': 'ubuntu',
       'filename': 'log/syslog.1',
       'text': (
           'Reporter <CRON> PID: 8442 (pam_unix(cron:session): session\n '
           'closed for user root)'),
       'timestamp': '2012-06-27 18:17:01',
       'timestamp_desc': definitions.TIME_DESCRIPTION_METADATA_MODIFICATION}]

  def _GetSheetRows(self, filename):
    """Parses the contents of the first sheet of an XLSX document.

    Args:
      filename (str): The file path of the XLSX document to parse.

    Returns:
      list[list[str]]: A list of lists representing the rows of the first sheet.

    Raises:
      ValueError: if the sheet cannot be found, or a string cannot be read.
    """
    rows = []
    with zipfile.ZipFile(filename) as zip_file:
      if self._SHEET1 not in zip_file.namelist():
        # Fail if we cannot find the expected first sheet.
        raise ValueError('Unable to locate expected sheet: {0:s}'.format(
            self._SHEET1))

      # Generate a reference table of shared strings if available.
      strings = []
      if self._SHARED_STRINGS in zip_file.namelist():
        with zip_file.open(self._SHARED_STRINGS) as zip_file_object:
          for _, element in ElementTree.iterparse(zip_file_object):
            if element.tag.endswith(self._SHARED_STRING_TAG):
              strings.append(element.text)

      row = []
      value = ''
      with zip_file.open(self._SHEET1) as zip_file_object:
        for _, element in ElementTree.iterparse(zip_file_object):
          if (element.tag.endswith(self._VALUE_STRING_TAG) or
              element.tag.endswith(self._SHARED_STRING_TAG)):
            value = element.text

          if element.tag.endswith(self._COLUMN_TAG):
            # Grab value from shared string reference table if type shared
            # string.
            if strings and element.attrib.get(
                self._TYPE_ATTRIBUTE) == self._SHARED_STRING_TYPE:
              try:
                value = strings[int(value)]
              except (IndexError, ValueError):
                raise ValueError(
                    'Unable to successfully dereference shared string.')

            row.append(value)

          # If we see the end tag of the row, record row in rows and reset.
          if element.tag.endswith(self._ROW_TAG):
            rows.append(row)
            row = []

    return rows

  def testGetFieldValues(self):
    """Tests the GetFieldValues function."""
    output_mediator = self._CreateOutputMediator()

    formatters_directory_path = self._GetTestFilePath(['formatters'])
    output_mediator.ReadMessageFormattersFromDirectory(
        formatters_directory_path)

    output_module = xlsx.XLSXOutputModule()

    event, event_data, event_data_stream = (
        containers_test_lib.CreateEventFromValues(self._TEST_EVENTS[0]))

    event_tag = events.EventTag()
    event_tag.AddLabels(['Malware', 'Printed'])

    expected_field_values = {
        'datetime': datetime.datetime(2012, 6, 27, 18, 17, 1),
        'display_name': '-',
        'message': (
            'Reporter <CRON> PID: 8442 (pam_unix(cron:session): session '
            'closed for user root)'),
        'parser': '-',
        'source': 'FILE',
        'source_long': 'Test log file',
        'tag': 'Malware Printed',
        'timestamp_desc': 'Metadata Modification Time'}

    field_values = output_module.GetFieldValues(
        output_mediator, event, event_data, event_data_stream, event_tag)

    self.assertEqual(field_values, expected_field_values)

  def testWriteFieldValues(self):
    """Tests the WriteFieldValues function."""
    output_mediator = self._CreateOutputMediator()

    formatters_directory_path = self._GetTestFilePath(['formatters'])
    output_mediator.ReadMessageFormattersFromDirectory(
        formatters_directory_path)

    output_module = xlsx.XLSXOutputModule()

    with shared_test_lib.TempDirectory() as temp_directory:
      xslx_file = os.path.join(temp_directory, 'xlsx.out')

      event, event_data, event_data_stream = (
          containers_test_lib.CreateEventFromValues(self._TEST_EVENTS[0]))

      event_tag = events.EventTag()
      event_tag.AddLabels(['Malware', 'Printed'])

      output_module.Open(path=xslx_file)

      try:
        output_module.WriteHeader(output_mediator)

        field_values = output_module.GetFieldValues(
            output_mediator, event, event_data, event_data_stream, event_tag)

        output_module.WriteFieldValues(output_mediator, field_values)

      finally:
        output_module.Close()

      try:
        rows = self._GetSheetRows(xslx_file)
      except ValueError as exception:
        self.fail(exception)

      expected_header = [
          'datetime', 'timestamp_desc', 'source', 'source_long',
          'message', 'parser', 'display_name', 'tag']
      expected_event_body = [
          '41087.76181712963', 'Metadata Modification Time', 'FILE',
          'Test log file',
          'Reporter <CRON> PID: 8442 (pam_unix(cron:session): session '
          'closed for user root)',
          '-', '-', 'Malware Printed']

      self.assertEqual(expected_header, rows[0])
      self.assertEqual(len(expected_event_body), len(rows[1]))
      self.assertEqual(expected_event_body, rows[1])

  def testWriteHeader(self):
    """Tests the WriteHeader function."""
    expected_header = [
        'datetime', 'timestamp_desc', 'source', 'source_long',
        'message', 'parser', 'display_name', 'tag']

    with shared_test_lib.TempDirectory() as temp_directory:
      output_mediator = self._CreateOutputMediator()
      output_module = xlsx.XLSXOutputModule()

      xlsx_file = os.path.join(temp_directory, 'xlsx.out')

      output_module.Open(path=xlsx_file)

      try:
        output_module.WriteHeader(output_mediator)
      finally:
        output_module.Close()

      try:
        rows = self._GetSheetRows(xlsx_file)
      except ValueError as exception:
        self.fail(exception)

      self.assertEqual(expected_header, rows[0])


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the Windows recycler parsers."""

import unittest

from plaso.parsers import recycler

from tests.parsers import test_lib


class WinRecycleBinParserTest(test_lib.ParserTestCase):
  """Tests for the Windows Recycle Bin parser."""

  def testParseVista(self):
    """Tests the Parse function on a Windows Vista RecycleBin file."""
    parser = recycler.WinRecycleBinParser()
    storage_writer = self._ParseFile(['recycler', '$II3DF3L.zip'], parser)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 1)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'data_type': 'windows:metadata:deleted_item',
        'deletion_time': '2012-03-12T20:49:58.6330000+00:00',
        'file_size': 724919,
        'original_filename': (
            'C:\\Users\\nfury\\Documents\\Alloy Research\\StarFury.zip')}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 0)
    self.CheckEventData(event_data, expected_event_values)

  def testParseWindows10(self):
    """Tests the Parse function on a Windows 10 RecycleBin file."""
    parser = recycler.WinRecycleBinParser()
    storage_writer = self._ParseFile(['recycler', '$I103S5F.jpg'], parser)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 1)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'data_type': 'windows:metadata:deleted_item',
        'deletion_time': '2016-06-29T21:37:45.6180000+00:00',
        'file_size': 222255,
        'original_filename': (
            'C:\\Users\\random\\Downloads\\bunnies.jpg')}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 0)
    self.CheckEventData(event_data, expected_event_values)


class WinRecyclerInfo2ParserTest(test_lib.ParserTestCase):
  """Tests for the Windows Recycler INFO2 parser."""

  def testParse(self):
    """Tests the Parse function on a Windows Recycler INFO2 file."""
    parser = recycler.WinRecyclerInfo2Parser()
    storage_writer = self._ParseFile(['recycler', 'INFO2'], parser)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 4)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'data_type': 'windows:metadata:deleted_item',
        'deletion_time': '2004-08-25T16:18:25.2370000+00:00',
        'drive_number': 2,
        'original_filename': (
            'C:\\Documents and Settings\\Mr. Evil\\Desktop\\lalsetup250.exe'),
        'record_index': 1}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 0)
    self.CheckEventData(event_data, expected_event_values)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the JSON-L parser plugin for Docker layer config files."""

import unittest

from plaso.parsers.jsonl_plugins import docker_layer_config

from tests.parsers.jsonl_plugins import test_lib


class DockerLayerLogJSONLPluginTest(test_lib.JSONLPluginTestCase):
  """Tests for the JSON-L parser plugin for Docker layer config files."""

  def testProcess(self):
    """Tests the Process function."""
    layer_identifier = (
        '3c9a9d7cc6a235eb2de58ca9ef3551c67ae42a991933ba4958d207b29142902b')
    path_segments = ['docker', 'graph', layer_identifier, 'json']

    plugin = docker_layer_config.DockerLayerConfigurationJSONLPlugin()
    storage_writer = self._ParseJSONLFileWithPlugin(path_segments, plugin)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 1)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'command': (
            '/bin/sh -c sed -i \'s/^#\\s*\\(deb.*universe\\)$/\\1/g\' '
            '/etc/apt/sources.list'),
        'creation_time': '2015-10-12T17:27:03.079273+00:00',
        'data_type': 'docker:layer:configuration',
        'layer_identifier': layer_identifier}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 0)
    self.CheckEventData(event_data, expected_event_values)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the .customDestinations-ms file parser."""

import unittest

from plaso.parsers import custom_destinations

from tests.parsers import test_lib


class CustomDestinationsParserTest(test_lib.ParserTestCase):
  """Tests for the .customDestinations-ms file parser."""

  def testParse(self):
    """Tests the Parse function."""
    parser = custom_destinations.CustomDestinationsParser()
    storage_writer = self._ParseFile(
        ['custom_destinations', '5afe4de1b92fc382.customDestinations-ms'],
        parser)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 45)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    # Test shortcut event data.
    expected_event_values = {
        'access_time': '2009-07-13T23:55:56.2481035+00:00',
        'command_line_arguments': (
            '{DE3895CB-077B-4C38-B6E3-F3DE1E0D84FC} %systemroot%\\\\'
            'system32\\\\control.exe /name Microsoft.Display'),
        'creation_time': '2009-07-13T23:55:56.2481035+00:00',
        'data_type': 'windows:lnk:link',
        'description': '@%systemroot%\\\\system32\\\\oobefldr.dll,-1262',
        'drive_serial_number': 0x24ba718b,
        'drive_type': 3,
        'env_var_location': '%SystemRoot%\\\\system32\\\\GettingStarted.exe',
        'file_attribute_flags': 0x00000020,
        'file_size': 11776,
        'icon_location': '%systemroot%\\\\system32\\\\display.dll',
        'link_target': (
            '<My Computer> C:\\\\Windows\\\\System32\\\\GettingStarted.exe'),
        'local_path': 'C:\\\\Windows\\\\System32\\\\GettingStarted.exe',
        'modification_time': '2009-07-14T01:39:11.3880000+00:00'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 43)
    self.CheckEventData(event_data, expected_event_values)

    # Test distributed link tracking event data.
    expected_event_values = {
        'creation_time': '2010-11-10T19:08:32.6562596+00:00',
        'data_type': 'windows:distributed_link_tracking:creation',
        'mac_address': '00:0c:29:03:1e:1e',
        'origin': '5afe4de1b92fc382.customDestinations-ms',
        'uuid': 'e9215b24-ecfd-11df-a81c-000c29031e1e'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 4)
    self.CheckEventData(event_data, expected_event_values)

    # Test shell item event data.
    expected_event_values = {
        'access_time': '2010-11-10T07:41:04+00:00',
        'creation_time': '2009-07-14T03:20:12+00:00',
        'data_type': 'windows:shell_item:file_entry',
        'file_reference': '2331-1',
        'long_name': 'System32',
        'modification_time': '2010-11-10T07:41:04+00:00',
        'name': 'System32',
        'origin': '5afe4de1b92fc382.customDestinations-ms',
        'shell_item_path': '<My Computer> C:\\\\Windows\\\\System32'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 41)
    self.CheckEventData(event_data, expected_event_values)

  def testParseWithEmpty(self):
    """Tests the Parse function with an empty jump list."""
    parser = custom_destinations.CustomDestinationsParser()
    storage_writer = self._ParseFile(
        ['custom_destinations', 'c98dce577f884ef8.customDestinations-ms'],
        parser)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

  def testParseWithComplex(self):
    """Tests the Parse function with a complex jump list."""
    parser = custom_destinations.CustomDestinationsParser()
    storage_writer = self._ParseFile(
        ['custom_destinations', '368d807282ccde9d.customDestinations-ms'],
        parser)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 9)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'access_time': '2024-01-16T06:12:42+00:00',
        'creation_time': '2023-07-12T18:11:20+00:00',
        'data_type': 'windows:shell_item:file_entry',
        'file_reference': '407354-78',
        'long_name': 'test',
        'modification_time': '2023-07-14T04:04:02+00:00',
        'name': 'test',
        'origin': '368d807282ccde9d.customDestinations-ms',
        'shell_item_path': '<My Computer> C:\\\\test'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 0)
    self.CheckEventData(event_data, expected_event_values)

    expected_event_values = {
        'access_time': '2024-01-16T06:12:41.2400523+00:00',
        'command_line_arguments': 'My Arguments',
        'creation_time': '2023-07-12T18:11:18.2749654+00:00',
        'data_type': 'windows:lnk:link',
        'description': None,
        'drive_serial_number': 0x2ca3d1ae,
        'drive_type': 3,
        'env_var_location': None,
        'file_attribute_flags': 0x00000010,
        'file_size': 4096,
        'icon_location': 'My Icon',
        'link_target': '<My Computer> C:\\\\test',
        'local_path': 'C:\\\\test',
        'modification_time': '2023-07-14T04:04:00.3349887+00:00'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 1)
    self.CheckEventData(event_data, expected_event_values)

    expected_event_values = {
        'creation_time': '2023-07-12T18:06:36.6282931+00:00',
        'data_type': 'windows:distributed_link_tracking:creation',
        'mac_address': '52:54:00:ee:b6:05',
        'origin': '368d807282ccde9d.customDestinations-ms',
        'uuid': 'd78dbcb3-20de-11ee-a2f8-525400eeb605'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 2)
    self.CheckEventData(event_data, expected_event_values)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""Output module for the log2timeline (L2T) CSV format.

For documentation on the L2T CSV format see:
  https://forensics.wiki/l2t_csv
"""

import datetime
import pytz

from acstore.containers import interface as containers_interface

from dfdatetime import interface as dfdatetime_interface
from dfdatetime import posix_time as dfdatetime_posix_time

from plaso.output import formatting_helper
from plaso.output import logger
from plaso.output import manager
from plaso.output import shared_dsv
from plaso.output import text_file


class L2TCSVEventFormattingHelper(shared_dsv.DSVEventFormattingHelper):
  """L2T CSV output module event formatting helper."""

  def GetFormattedMACBGroup(self, output_mediator, macb_group):
    """Retrieves a string representation of a MACB group.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      macb_group (list[tuple[event, event_data, event_data_stream, event_tag]]):
          group of event, event_data, event_data_stream and event_tag objects
          with identical timestamps, attributes and values.

    Returns:
      str: string representation of the MACB group.
    """
    timestamp_descriptions = [
        event.timestamp_desc for event, _, _, _ in macb_group]

    field_values = []
    for field_name in self._field_names:
      if field_name == 'MACB':
        field_value = output_mediator.GetMACBRepresentationFromDescriptions(
            timestamp_descriptions)
      elif field_name == 'type':
        # TODO: fix timestamp description in source.
        field_value = '; '.join(timestamp_descriptions)
      else:
        event, event_data, event_data_stream, event_tag = macb_group[0]
        field_value = self._field_formatting_helper.GetFormattedField(
            output_mediator, field_name, event, event_data, event_data_stream,
            event_tag)

      if field_value is None:
        field_value = '-'

      field_value = self._SanitizeField(field_value)
      field_values.append(field_value)

    return self.field_delimiter.join(field_values)


class L2TCSVFieldFormattingHelper(formatting_helper.FieldFormattingHelper):
  """L2T CSV output module field formatting helper."""

  # Maps the name of a fields to a a callback function that formats
  # the field value.
  _FIELD_FORMAT_CALLBACKS = {
      'date': '_FormatDate',
      'desc': '_FormatMessage',
      'extra': '_FormatExtraAttributes',
      'filename': '_FormatDisplayName',
      'format': '_FormatParser',
      'host': '_FormatHostname',
      'inode': '_FormatInode',
      'MACB': '_FormatMACB',
      'notes': '_FormatTag',
      'short': '_FormatMessageShort',
      'source': '_FormatSourceShort',
      'sourcetype': '_FormatSource',
      'time': '_FormatTime',
      'timezone': '_FormatTimeZone',
      'type': '_FormatType',
      'user': '_FormatUsername',
      'values': '_FormatValues',
      'version': '_FormatVersion'}

  # Note that L2T CSV defines certain fields as part of the format.
  _RESERVED_VARIABLE_NAMES = frozenset([
      '_event_values_hash',
      '_parser_chain',
      'body',
      'data_type',
      'date_time',
      'display_name',
      'filename',
      'hostname',
      'http_headers',
      'inode',
      'mapped_files',
      'metadata',
      'offset',
      'path_spec',
      'query',
      'source_long',
      'source_short',
      'tag',
      'timestamp',
      'timestamp_desc',
      'timezone',
      'username'])

  # The field format callback methods require specific arguments hence
  # the check for unused arguments is disabled here.
  # pylint: disable=unused-argument

  def _FormatDate(self, output_mediator, event, event_data, event_data_stream):
    """Formats a date field.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.

    Returns:
      str: date formatted as "MM/DD/YYYY" or "00/00/0000" on error.
    """
    # For now check if event.timestamp is set, to mimic existing behavior of
    # using 00/00/0000 for 0 timestamp values.
    if not event.timestamp:
      return '00/00/0000'

    date_time = event.date_time
    if not date_time or date_time.is_local_time:
      date_time = dfdatetime_posix_time.PosixTimeInMicroseconds(
          timestamp=event.timestamp)

    # Note that GetDateWithTimeOfDay will return the date and time in UTC,
    # so no adjustment for date_time.time_zone_offset is needed.
    year, month, day_of_month, hours, minutes, seconds = (
        date_time.GetDateWithTimeOfDay())

    if output_mediator.time_zone != pytz.UTC:
      try:
        datetime_object = datetime.datetime(
            year, month, day_of_month, hours, minutes, seconds,
            tzinfo=pytz.UTC)

        datetime_object = datetime_object.astimezone(output_mediator.time_zone)

        year = datetime_object.year
        month = datetime_object.month
        day_of_month = datetime_object.day

      except (OSError, OverflowError, TypeError, ValueError):
        year, month, day_of_month = (None, None, None)

    if None in (year, month, day_of_month):
      self._ReportEventError(event, event_data, (
          'unable to copy timestamp: {0!s} to a human readable date. '
          'Defaulting to: "00/00/0000"').format(event.timestamp))
      return '00/00/0000'

    return '{0:02d}/{1:02d}/{2:04d}'.format(month, day_of_month, year)

  def _FormatExtraAttributes(
      self, output_mediator, event, event_data, event_data_stream):
    """Formats an extra attributes field.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.

    Returns:
      str: extra attributes field.
    """
    message_formatter = output_mediator.GetMessageFormatter(
        event_data.data_type)
    if not message_formatter:
      logger.warning(
          'Using default message formatter for data type: {0:s}'.format(
              event_data.data_type))
      message_formatter = self._DEFAULT_MESSAGE_FORMATTER

    formatted_attribute_names = (
        message_formatter.GetFormatStringAttributeNames())
    formatted_attribute_names.update(self._RESERVED_VARIABLE_NAMES)

    extra_attributes = []
    for attribute_name, attribute_value in event_data.GetAttributes():
      if attribute_name in formatted_attribute_names:
        continue

      # Ignore attribute container identifier and date and time values.
      if isinstance(attribute_value, (
          containers_interface.AttributeContainerIdentifier,
          dfdatetime_interface.DateTimeValues)):
        continue

      if (isinstance(attribute_value, list) and attribute_value and
          isinstance(attribute_value[0],
                     dfdatetime_interface.DateTimeValues)):
        continue

      # Some parsers have written bytes values to storage.
      if isinstance(attribute_value, bytes):
        attribute_value = attribute_value.decode('utf-8', 'replace')
        logger.warning(
            'Found bytes value for attribute "{0:s}" for data type: '
            '{1!s}. Value was converted to UTF-8: "{2:s}"'.format(
                attribute_name, event_data.data_type, attribute_value))

      # With ! in {1!s} we force a string conversion since some of
      # the extra attributes values can be integer, float point or
      # boolean values.
      extra_attributes.append('{0:s}: {1!s}'.format(
          attribute_name, attribute_value))

    if event_data_stream:
      for attribute_name, attribute_value in event_data_stream.GetAttributes():
        if attribute_name != 'path_spec':
          extra_attributes.append('{0:s}: {1!s}'.format(
              attribute_name, attribute_value))

    extra_attributes = '; '.join(sorted(extra_attributes))

    return extra_attributes.replace('\n', '-').replace('\r', '')

  def _FormatType(self, output_mediator, event, event_data, event_data_stream):
    """Formats a type field.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.

    Returns:
      str: type field.
    """
    return getattr(event, 'timestamp_desc', '-')

  def _FormatVersion(
      self, output_mediator, event, event_data, event_data_stream):
    """Formats a version field.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.

    Returns:
      str: version field.
    """
    return '2'

  # pylint: enable=unused-argument


class L2TCSVOutputModule(text_file.SortedTextFileOutputModule):
  """CSV format used by log2timeline, with 17 fixed fields."""

  NAME = 'l2tcsv'
  DESCRIPTION = 'CSV format used by legacy log2timeline, with 17 fixed fields.'

  _FIELD_NAMES = [
      'date', 'time', 'timezone', 'MACB', 'source', 'sourcetype', 'type',
      'user', 'host', 'short', 'desc', 'version', 'filename', 'inode', 'notes',
      'format', 'extra']

  _SORT_KEY_FIELD_NAMES = ['time', 'filename', 'inode']

  def __init__(self):
    """Initializes an output module."""
    field_formatting_helper = L2TCSVFieldFormattingHelper()
    event_formatting_helper = L2TCSVEventFormattingHelper(
        field_formatting_helper, self._FIELD_NAMES)
    super(L2TCSVOutputModule, self).__init__(event_formatting_helper)

  def _GetString(self, output_mediator, field_values):
    """Retrieves an output string.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      field_values (dict[str, str]): output field values per name.

    Returns:
      str: output string.
    """
    output_text = self._event_formatting_helper.field_delimiter.join(
        field_values.values())
    return ''.join([output_text, '\n'])

  def WriteFieldValuesOfMACBGroup(self, output_mediator, macb_group):
    """Writes field values of a MACB group to the output.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
      macb_group (list[tuple[event, event_data, event_data_stream, event_tag]]):
          group of event, event_data, event_data_stream and event_tag objects
          with identical timestamps, attributes and values.
    """
    output_text = self._event_formatting_helper.GetFormattedMACBGroup(
        output_mediator, macb_group)

    self.WriteLine(output_text)

  def WriteHeader(self, output_mediator):
    """Writes the header to the output.

    Args:
      output_mediator (OutputMediator): mediates interactions between output
          modules and other components, such as storage and dfVFS.
    """
    output_text = self._event_formatting_helper.GetFormattedFieldNames()
    self.WriteLine(output_text)


manager.OutputManager.RegisterOutput(L2TCSVOutputModule)



# -*- coding: utf-8 -*-
"""Windows Registry custom event formatter helpers."""

from plaso.formatters import interface
from plaso.formatters import manager


class WindowsRegistryValuesFormatterHelper(
    interface.CustomEventFormatterHelper):
  """Windows Registry values formatter helper."""

  IDENTIFIER = 'windows_registry_values'

  def FormatEventValues(self, output_mediator, event_values):
    """Formats event values using the helper.

    Args:
      output_mediator (OutputMediator): output mediator.
      event_values (dict[str, object]): event values.
    """
    values = event_values.get('values', None)
    if not values:
      values_string = '(empty)'

    elif isinstance(values, list):
      value_strings = []
      for name, data_type, data in sorted(values):
        if not name:
          name = '(default)'
        if not data:
          data = '(empty)'
        elif isinstance(data, bytes):
          data = '({0:d} bytes)'.format(len(data))

        value_strings.append(
            '{0:s}: [{1:s}] {2:s}'.format(name, data_type, data))

      values_string = ' '.join(value_strings)

    else:
      values_string = values

    event_values['values'] = values_string


manager.FormattersManager.RegisterEventFormatterHelper(
    WindowsRegistryValuesFormatterHelper)



# -*- coding: utf-8 -*-
"""Artifact attribute containers."""

from acstore.containers import interface
from acstore.containers import manager

from plaso.lib import definitions


class ArtifactAttributeContainer(interface.AttributeContainer):
  """Base class to represent an artifact attribute container."""


class EnvironmentVariableArtifact(ArtifactAttributeContainer):
  """Environment variable artifact attribute container.

  Also see:
    https://en.wikipedia.org/wiki/Environment_variable

  Attributes:
    case_sensitive (bool): True if environment variable name is case sensitive.
    name (str): environment variable name such as "SystemRoot" as in
        "%SystemRoot%" or "HOME" as in "$HOME".
    value (str): environment variable value such as "C:\\Windows" or
        "/home/user".
  """

  CONTAINER_TYPE = 'environment_variable'

  SCHEMA = {
      'case_sensitive': 'bool',
      'name': 'str',
      'value': 'str'}

  def __init__(self, case_sensitive=True, name=None, value=None):
    """Initializes an environment variable artifact.

    Args:
      case_sensitive (Optional[bool]): True if environment variable name
          is case sensitive.
      name (Optional[str]): environment variable name.
      value (Optional[str]): environment variable value.
    """
    super(EnvironmentVariableArtifact, self).__init__()
    self.case_sensitive = case_sensitive
    self.name = name
    self.value = value


class HostnameArtifact(ArtifactAttributeContainer):
  """Hostname artifact attribute container.

  Also see:
    https://en.wikipedia.org/wiki/Hostname
    Cybox / Stix Hostname Object

  Attributes:
    name (str): name of the host according to the naming schema.
    schema (str): naming schema such as "DNS", "NIS", "SMB/NetBIOS".
  """

  CONTAINER_TYPE = 'hostname'

  SCHEMA = {
      'name': 'str',
      'schema': 'str'}

  def __init__(self, name=None, schema='DNS'):
    """Initializes a hostname artifact.

    Args:
      name (Optional[str]): name of the host according to the naming schema.
      schema (Optional[str]): naming schema.
    """
    super(HostnameArtifact, self).__init__()
    self.name = name
    self.schema = schema


class OperatingSystemArtifact(ArtifactAttributeContainer):
  """Operating system artifact attribute container.

  Attributes:
    family (str): operating system family name, such as "Linux", "MacOS"
        or "Windows", defined in definitions.OPERATING_SYSTEM_FAMILIES.
        This value is used to programmatically link a parser preset to an
        operating system and therefore must be one of predefined values.
    name (str): operating system name, such as "macOS Mojave" or "Windows XP".
        This value is used to programmatically link a parser preset to an
        operating system and therefore must be one of predefined values.
    product (str): product information, such as "macOS Mojave" or "Windows
        Professional XP". This value is typically obtained from the source data.
    version (str): version, such as "10.14.1" or "5.1". This value is typically
        obtained from the source data.
  """

  CONTAINER_TYPE = 'operating_system'

  SCHEMA = {
      'family': 'str',
      'name': 'str',
      'product': 'str',
      'version': 'str'}

  _DEFAULT_FAMILY_AND_VERSION = (
      definitions.OPERATING_SYSTEM_FAMILY_UNKNOWN, (0, 0))

  _FAMILY_AND_VERSION_PER_NAME = {
      'Windows 2000': (definitions.OPERATING_SYSTEM_FAMILY_WINDOWS_NT, (5, 0)),
      'Windows 2003': (definitions.OPERATING_SYSTEM_FAMILY_WINDOWS_NT, (5, 2)),
      'Windows 2003 R2': (
          definitions.OPERATING_SYSTEM_FAMILY_WINDOWS_NT, (5, 2)),
      'Windows 2008': (definitions.OPERATING_SYSTEM_FAMILY_WINDOWS_NT, (6, 0)),
      'Windows 2008 R2': (
          definitions.OPERATING_SYSTEM_FAMILY_WINDOWS_NT, (6, 1)),
      'Windows 2012': (definitions.OPERATING_SYSTEM_FAMILY_WINDOWS_NT, (6, 2)),
      'Windows 2012 R2': (
          definitions.OPERATING_SYSTEM_FAMILY_WINDOWS_NT, (6, 3)),
      'Windows 2016': (definitions.OPERATING_SYSTEM_FAMILY_WINDOWS_NT, (10, 0)),
      'Windows 2019': (definitions.OPERATING_SYSTEM_FAMILY_WINDOWS_NT, (10, 0)),
      'Windows 7': (definitions.OPERATING_SYSTEM_FAMILY_WINDOWS_NT, (6, 1)),
      'Windows 8': (definitions.OPERATING_SYSTEM_FAMILY_WINDOWS_NT, (6, 2)),
      'Windows 8.1': (definitions.OPERATING_SYSTEM_FAMILY_WINDOWS_NT, (6, 3)),
      'Windows 10': (definitions.OPERATING_SYSTEM_FAMILY_WINDOWS_NT, (10, 0)),
      'Windows Vista': (definitions.OPERATING_SYSTEM_FAMILY_WINDOWS_NT, (6, 0)),
      'Windows XP': (definitions.OPERATING_SYSTEM_FAMILY_WINDOWS_NT, (5, 1))}

  def __init__(self, family=None, product=None, version=None):
    """Initializes an operating system artifact.

    Args:
      family (Optional[str]): operating system family name, such as "Linux",
        "MacOS" or "Windows", defined in definitions.OPERATING_SYSTEM_FAMILIES.
        This value is used to programmatically link a parser preset to an
        operating system and therefore must be one of predefined values.
      product (Optional[str]): product information, such as "macOS Mojave" or
          "Windows Professional XP". This value is typically obtained from the
          source data.
      version (Optional[str]): version, such as "10.14.1" or "5.1". This value
          is typically obtained from the source data.
    """
    super(OperatingSystemArtifact, self).__init__()
    self.family = family
    self.name = None
    self.product = product
    self.version = version

    if product:
      self.name = self._GetNameFromProduct()

  @property
  def version_tuple(self):
    """Tuple[int]: version tuple or None if version is not set or invalid."""
    try:
      # pylint: disable=consider-using-generator
      return tuple([int(digit, 10) for digit in self.version.split('.')])
    except (AttributeError, TypeError, ValueError):
      return None

  def _GetNameFromProduct(self):
    """Determines the predefined operating system name from the product.

    Returns:
      str: operating system name, such as "macOS Mojave" or "Windows XP" or
          None if the name cannot be determined. This value is used to
          programmatically link a parser preset to an operating system and
          therefore must be one of predefined values.
    """
    product = self.product or ''
    product = product.split(' ')
    product_lower_case = [segment.lower() for segment in product]
    number_of_segments = len(product)

    if 'windows' in product_lower_case:
      segment_index = product_lower_case.index('windows') + 1
      if product_lower_case[segment_index] in ('(r)', 'server', 'web'):
        segment_index += 1

      # Check if the version has a suffix.
      suffix_segment_index = segment_index + 1
      if (suffix_segment_index < number_of_segments and
          product_lower_case[suffix_segment_index] == 'r2'):
        return 'Windows {0:s} R2'.format(product[segment_index])

      return 'Windows {0:s}'.format(product[segment_index])

    return None

  def IsEquivalent(self, other):
    """Determines if 2 operating system artifacts are equivalent.

    This function compares the operating systems based in order of:
    * name derived from product
    * family and version
    * family

    Args:
      other (OperatingSystemArtifact): operating system artifact attribute
          container to compare with.

    Returns:
      bool: True if the operating systems are considered equivalent, False if
          the most specific criteria do no match, or no criteria are available.
    """
    if self.name and other.name:
      return self.name == other.name

    if self.name:
      self_family, self_version_tuple = self._FAMILY_AND_VERSION_PER_NAME.get(
          self.name, self._DEFAULT_FAMILY_AND_VERSION)
      return (
          self_family == other.family and
          self_version_tuple == other.version_tuple)

    if self.family and self.version:
      if other.name:
        other_family, other_version_tuple = (
            self._FAMILY_AND_VERSION_PER_NAME.get(
                other.name, self._DEFAULT_FAMILY_AND_VERSION))
      else:
        other_family = other.family
        other_version_tuple = other.version_tuple

      return (
          self.family == other_family and
          self.version_tuple == other_version_tuple)

    if self.family:
      if other.name:
        other_family, _ = self._FAMILY_AND_VERSION_PER_NAME.get(
            other.name, self._DEFAULT_FAMILY_AND_VERSION)
      else:
        other_family = other.family

      return self.family == other_family

    return False


class PathArtifact(ArtifactAttributeContainer):
  """Path artifact attribute container.

  Attributes:
    data_stream (str): name of a data stream.
    path_segment_separator (str): path segment separator.
    path_segments (list[str]): path segments.
  """

  CONTAINER_TYPE = 'path'

  SCHEMA = {
      'data_stream': 'str',
      'path_segment_separator': 'str',
      'path_segments': 'List[str]'}

  def __init__(self, data_stream=None, path=None, path_segment_separator='/'):
    """Initializes a path artifact.

    Args:
      data_stream (Optional[str]): name of a data stream.
      path (Optional[str]): a path.
      path_segment_separator (Optional[str]): path segment separator.
    """
    super(PathArtifact, self).__init__()
    self.data_stream = data_stream
    self.path_segment_separator = path_segment_separator
    self.path_segments = self._SplitPath(path, path_segment_separator)

  def __eq__(self, other):
    """Determines if the path is equal to other.

    Args:
      other (str): path to compare against.

    Returns:
      bool: True if the path are equal to other.
    """
    if not isinstance(other, str):
      return False

    other_path_segments = self._SplitPath(other, self.path_segment_separator)
    return self.path_segments == other_path_segments

  def __ge__(self, other):
    """Determines if the path are greater than or equal to other.

    Args:
      other (str): path to compare against.

    Returns:
      bool: True if the path are greater than or equal to other.

    Raises:
      ValueError: if other is not an instance of string.
    """
    if not isinstance(other, str):
      raise ValueError('Other not an instance of string.')

    other_path_segments = self._SplitPath(other, self.path_segment_separator)
    return self.path_segments >= other_path_segments

  def __gt__(self, other):
    """Determines if the path are greater than other.

    Args:
      other (str): path to compare against.

    Returns:
      bool: True if the path are greater than other.

    Raises:
      ValueError: if other is not an instance of string.
    """
    if not isinstance(other, str):
      raise ValueError('Other not an instance of string.')

    other_path_segments = self._SplitPath(other, self.path_segment_separator)
    return self.path_segments > other_path_segments

  def __le__(self, other):
    """Determines if the path are greater than or equal to other.

    Args:
      other (str): path to compare against.

    Returns:
      bool: True if the path are greater than or equal to other.

    Raises:
      ValueError: if other is not an instance of string.
    """
    if not isinstance(other, str):
      raise ValueError('Other not an instance of string.')

    other_path_segments = self._SplitPath(other, self.path_segment_separator)
    return self.path_segments <= other_path_segments

  def __lt__(self, other):
    """Determines if the path are less than other.

    Args:
      other (str): path to compare against.

    Returns:
      bool: True if the path are less than other.

    Raises:
      ValueError: if other is not an instance of string.
    """
    if not isinstance(other, str):
      raise ValueError('Other not an instance of string.')

    other_path_segments = self._SplitPath(other, self.path_segment_separator)
    return self.path_segments < other_path_segments

  def __ne__(self, other):
    """Determines if the path are not equal to other.

    Args:
      other (str): path to compare against.

    Returns:
      bool: True if the path are not equal to other.
    """
    if not isinstance(other, str):
      return False

    other_path_segments = self._SplitPath(other, self.path_segment_separator)
    return self.path_segments != other_path_segments

  def _SplitPath(self, path, path_segment_separator):
    """Splits a path.

    Args:
      path (str): a path.
      path_segment_separator (str): path segment separator.

    Returns:
      list[str]: path segments.
    """
    path = path or ''
    split_path = path.split(path_segment_separator)

    path_segments = [split_path[0]]
    path_segments.extend(list(filter(None, split_path[1:])))

    return path_segments

  def ContainedIn(self, other):
    """Determines if the path are contained in other.

    Args:
      other (str): path to compare against.

    Returns:
      bool: True if the path is contained in other.
    """
    if isinstance(other, str):
      number_of_path_segments = len(self.path_segments)
      other_path_segments = self._SplitPath(other, self.path_segment_separator)
      number_of_other_path_segments = len(other_path_segments)
      if number_of_path_segments < number_of_other_path_segments:
        maximum_compare_length = (
            number_of_other_path_segments - number_of_path_segments + 1)

        for compare_start_index in range(0, maximum_compare_length):
          compare_end_index = compare_start_index + number_of_path_segments
          compare_path_segments = other_path_segments[
              compare_start_index:compare_end_index]
          if self.path_segments == compare_path_segments:
            return True

    return False


class SourceConfigurationArtifact(ArtifactAttributeContainer):
  """Source configuration artifact attribute container.

  The source configuration contains the configuration data of a source
  that is (or going to be) processed such storage media image or a mounted
  directory.

  Attributes:
    path (str): path of the source.
    source_type (str): type of source.
  """

  CONTAINER_TYPE = 'source_configuration'

  SCHEMA = {
      'path': 'str',
      'source_type': 'str'}

  def __init__(self, path=None, source_type=None):
    """Initializes a source configuration artifact.

    Args:
      path (Optional[str]): path of the source.
      source_type (Optional[str]): type of source.
    """
    super(SourceConfigurationArtifact, self).__init__()
    self.path = path
    self.source_type = source_type


class SystemConfigurationArtifact(ArtifactAttributeContainer):
  """System configuration artifact attribute container.

  The system configuration contains the configuration data of a specific
  system installation such as Windows or Linux.

  Attributes:
    available_time_zones (list[TimeZone]): available time zones.
    code_page (str): system code page.
    environment_variables (list[EnvironmentVariableArtifact]): environment
        variables.
    hostname (HostnameArtifact): hostname.
    keyboard_layout (str): keyboard layout.
    language (str): system language.
    operating_system (str): operating system for example "MacOS" or "Windows".
    operating_system_product (str): operating system product for example
        "Windows XP".
    operating_system_version (str): operating system version for example
        "10.9.2" or "8.1".
    path_specs (list[dfvfs.PathSpec]): path specifications of the file systems
        the system configuration was retrieved from.
    time_zone (str): system time zone.
    user_accounts (list[UserAccountArtifact]): user accounts.
  """

  CONTAINER_TYPE = 'system_configuration'

  # TODO: add SCHEMA
  # SCHEMA = {
  #     'code_page': 'str',
  #     'environment_variables': Llist[EnvironmentVariableArtifact]',
  #     'keyboard_layout': 'str',
  #     'hostname': 'HostnameArtifact',
  #     'language': 'str',
  #     'operating_system': 'str',
  #     'operating_system_product': 'str',
  #     'operating_system_version': 'str',
  #     'path_specs': 'List[dfvfs.PathSpec]',
  #     'time_zone': 'str'}

  def __init__(self, code_page=None, language=None, time_zone=None):
    """Initializes a system configuration artifact.

    Args:
      code_page (Optional[str]): system code page.
      language (Optional[str]): system language.
      time_zone (Optional[str]): system time zone.
    """
    super(SystemConfigurationArtifact, self).__init__()
    # TODO: kept for backwards compatibility.
    self.available_time_zones = []
    self.code_page = code_page
    self.environment_variables = []
    self.hostname = None
    self.keyboard_layout = None
    self.language = language
    self.operating_system = None
    self.operating_system_product = None
    self.operating_system_version = None
    self.path_specs = []
    self.time_zone = time_zone
    # TODO: kept for backwards compatibility.
    self.user_accounts = []


class TimeZoneArtifact(ArtifactAttributeContainer):
  """Time zone artifact attribute container.

  Attributes:
    localized_name (str): name describing the time zone in localized language
        for example "Greenwich (standaardtijd)".
    mui_form (str): MUI form of the name describing the time zone for example
        "@tzres.dll,-112".
    name (str): name describing the time zone for example "Greenwich Standard
        Time".
    offset (int): time zone offset in number of minutes from UTC.
  """

  CONTAINER_TYPE = 'time_zone'

  SCHEMA = {
      'localized_name': 'str',
      'mui_form': 'str',
      'name': 'str',
      'offset': 'int'}

  def __init__(
      self, localized_name=None, mui_form=None, name=None, offset=None):
    """Initializes a time zone artifact.

    Args:
      localized_name (Optional[str]): name describing the time zone in localized
          language for example "Greenwich (standaardtijd)".
      mui_form (Optional[str]): MUI form of the name describing the time zone
          for example "@tzres.dll,-112".
      name (Optional[str]): name describing the time zone for example "Greenwich
          Standard Time".
      offset (Optional[int]): time zone offset in number of minutes from UTC.
    """
    super(TimeZoneArtifact, self).__init__()
    self.localized_name = localized_name
    self.mui_form = mui_form
    self.name = name
    self.offset = offset


class UserAccountArtifact(ArtifactAttributeContainer):
  """User account artifact attribute container.

  Also see:
    Cybox / Stix User Account Object

  Attributes:
    full_name (str): name describing the user.
    group_identifier (str): identifier of the primary group the user is part of.
    identifier (str): user identifier.
    user_directory (str): path of the user (or home or profile) directory.
    username (str): name uniquely identifying the user.
  """

  CONTAINER_TYPE = 'user_account'

  SCHEMA = {
      'full_name': 'str',
      'group_identifier': 'str',
      'identifier': 'str',
      'user_directory': 'str',
      'username': 'str'}

  def __init__(
      self, full_name=None, group_identifier=None, identifier=None,
      path_separator='/', user_directory=None, username=None):
    """Initializes a user account artifact.

    Args:
      full_name (Optional[str]): name describing the user.
      group_identifier (Optional[str]): identifier of the primary group
          the user is part of.
      identifier (Optional[str]): user identifier.
      path_separator (Optional[str]): path segment separator.
      user_directory (Optional[str]): path of the user (or home or profile)
          directory.
      username (Optional[str]): name uniquely identifying the user.
    """
    super(UserAccountArtifact, self).__init__()
    self._path_separator = path_separator
    self.full_name = full_name
    self.group_identifier = group_identifier
    self.identifier = identifier
    # TODO: add shell.
    self.user_directory = user_directory
    self.username = username

  def GetUserDirectoryPathSegments(self):
    """Retrieves the path segments of the user directory.

    Returns:
      list[str]: path segments of the user directory or an empty list if no
          user directory is set.
    """
    if not self.user_directory:
      return []
    return self.user_directory.split(self._path_separator)


class WindowsEventLogMessageFileArtifact(ArtifactAttributeContainer):
  """Windows EventLog message file artifact attribute container.

  Attributes:
    path (str): path.
    windows_path (str): path as defined by the Windows EventLog provider.
  """

  CONTAINER_TYPE = 'windows_eventlog_message_file'

  SCHEMA = {
      'path': 'str',
      'windows_path': 'str'}

  def __init__(self, path=None, windows_path=None):
    """Initializes a Windows EventLog message file artifact.

    Args:
      path (Optional[str]): path.
      windows_path (Optional[str]): path as defined by the Window EventLog
          provider.
    """
    super(WindowsEventLogMessageFileArtifact, self).__init__()
    self.path = path
    self.windows_path = windows_path


class WindowsEventLogMessageStringArtifact(ArtifactAttributeContainer):
  """Windows EventLog message string artifact attribute container.

  Attributes:
    language_identifier (str): language identifier.
    message_identifier (int): message identifier.
    string (str): string.
  """

  CONTAINER_TYPE = 'windows_eventlog_message_string'

  SCHEMA = {
      '_message_file_identifier': 'AttributeContainerIdentifier',
      'language_identifier': 'int',
      'message_identifier': 'int',
      'string': 'str'}

  _SERIALIZABLE_PROTECTED_ATTRIBUTES = [
      '_message_file_identifier']

  def __init__(
      self, language_identifier=None, message_identifier=None, string=None):
    """Initializes a Windows EventLog message string artifact.

    Args:
      language_identifier (Optional[str]): language identifier.
      message_identifier (Optional[int]): message identifier.
      string (Optional[str]): string.
    """
    super(WindowsEventLogMessageStringArtifact, self).__init__()
    self._message_file_identifier = None
    self.language_identifier = language_identifier
    self.message_identifier = message_identifier
    self.string = string

  def GetMessageFileIdentifier(self):
    """Retrieves the identifier of the associated message file.

    Returns:
      AttributeContainerIdentifier: message file identifier or None when
          not set.
    """
    return self._message_file_identifier

  def SetMessageFileIdentifier(self, message_file_identifier):
    """Sets the identifier of the associated message file.

    Args:
      message_file_identifier (AttributeContainerIdentifier): message file
          identifier.
    """
    self._message_file_identifier = message_file_identifier


class WindowsEventLogProviderArtifact(ArtifactAttributeContainer):
  """Windows EventLog provider artifact attribute container.

  Attributes:
    additional_identifier (str): additional identifier of the provider,
        contains a GUID.
    category_message_files (list[str]): filenames of the category message files.
    event_message_files (list[str]): filenames of the event message files.
    identifier (str): identifier of the provider, contains a GUID.
    log_sources (list[str]): names of the corresponding Event Log sources.
    log_types (list[str]): Windows Event Log types.
    parameter_message_files (list[str]): filenames of the parameter message
        files.
  """

  CONTAINER_TYPE = 'windows_eventlog_provider'

  SCHEMA = {
      '_system_configuration_identifier': 'AttributeContainerIdentifier',
      'additional_identifier': 'str',
      'category_message_files': 'List[str]',
      'event_message_files': 'List[str]',
      'identifier': 'str',
      'log_sources': 'List[str]',
      'log_types': 'List[str]',
      'parameter_message_files': 'List[str]'}

  _SERIALIZABLE_PROTECTED_ATTRIBUTES = [
      '_system_configuration_identifier']

  def __init__(
      self, category_message_files=None, event_message_files=None,
      identifier=None, log_source=None, log_type=None,
      parameter_message_files=None):
    """Initializes a Windows EventLog provider artifact.

    Args:
      category_message_files (Optional[list[str]]): filenames of the category
          message files.
      event_message_files (Optional[list[str]]): filenames of the event message
          files.
      identifier (Optional[str]): identifier of the provider, contains a GUID.
      log_source (Optional[str]): name of the Windows EventLog source.
      log_type (Optional[str]): Windows EventLog type.
      parameter_message_files (Optional[list[str]]): filenames of the parameter
          message files.
    """
    super(WindowsEventLogProviderArtifact, self).__init__()
    self._system_configuration_identifier = None
    self.additional_identifier = None
    self.category_message_files = category_message_files or []
    self.event_message_files = event_message_files or []
    self.identifier = identifier
    self.log_sources = []
    self.log_types = []
    self.parameter_message_files = parameter_message_files or []

    if log_source:
      self.log_sources.append(log_source)
    if log_type:
      self.log_types.append(log_type)


class WindowsMountedDeviceArtifact(ArtifactAttributeContainer):
  """Windows mounted device artifact attribute container.

  Attributes:
    device (str): device.
    disk_identity (int): MBR disk identity.
    identifier (str): identifier.
    partition_identifier (str): GPT partition identifier.
    partition_offset (int): MBR partition offset.
  """

  CONTAINER_TYPE = 'windows_mounted_device'

  SCHEMA = {
      'device': 'str',
      'disk_identity': 'int',
      'identifier': 'str',
      'partition_identifier': 'str',
      'partition_offset': 'int'}

  def __init__(self, identifier=None):
    """Initializes a Windows mounted device artifact.

    Args:
      identifier (Optional[str]): identifier.
    """
    super(WindowsMountedDeviceArtifact, self).__init__()
    self.device = None
    self.disk_identity = None
    self.identifier = identifier
    self.partition_identifier = None
    self.partition_offset = None


class WindowsServiceConfigurationArtifact(ArtifactAttributeContainer):
  """Windows service (or driver) configuration artifact attribute container.

  Attributes:
    error_control (int): error control value of the service (or driver)
        executable.
    image_path (str): path of the service (or driver) executable.
    name (str): name of the service (or driver).
    object_name (str): service object name.
    service_dll (str): service DLL.
    service_type (int): service (or driver) type.
    start_type (int): service (or driver) start type.
  """

  CONTAINER_TYPE = 'windows_service_configuration'

  SCHEMA = {
      'error_control': 'int',
      'image_path': 'str',
      'name': 'str',
      'object_name': 'str',
      'service_dll': 'str',
      'service_type': 'int',
      'start_type': 'int'}

  def __init__(self, name=None, service_type=None, start_type=None):
    """Initializes a Windows service (or driver) configuration artifact.

    Args:
      name (Optional[str]): name of the service (or driver).
      service_type (Optional[int]): service (or driver) type.
      start_type (Optional[int]): service (or driver) start type.
    """
    super(WindowsServiceConfigurationArtifact, self).__init__()
    self.error_control = None
    self.image_path = None
    self.name = name
    self.object_name = None
    self.service_dll = None
    self.service_type = service_type
    self.start_type = start_type


class WindowsWevtTemplateEvent(ArtifactAttributeContainer):
  """Windows WEVT_TEMPLATE event definition.

  Attributes:
    identifier (int): event identifier.
    message_identifier (int): identifier of the event message.
    provider_identifier (str): identifier of the EventLog provider, contains
        a GUID.
    version (int): event version.
  """

  CONTAINER_TYPE = 'windows_wevt_template_event'

  SCHEMA = {
      '_message_file_identifier': 'AttributeContainerIdentifier',
      'identifier': 'int',
      'message_identifier': 'int',
      'provider_identifier': 'str',
      'version': 'int'}

  _SERIALIZABLE_PROTECTED_ATTRIBUTES = [
      '_message_file_identifier']

  def __init__(
      self, identifier=None, message_identifier=None, provider_identifier=None,
      version=None):
    """Initializes a Windows WEVT_TEMPLATE event definition artifact.

    Args:
      identifier (Optional[int]): event identifier.
      message_identifier (Optional[int]): identifier of the event message.
      provider_identifier (Optional[str]): identifier of the EventLog provider,
          contains a GUID.
      version (Optional[int]): event version.
    """
    super(WindowsWevtTemplateEvent, self).__init__()
    self._message_file_identifier = None
    self.identifier = identifier
    self.message_identifier = message_identifier
    self.provider_identifier = provider_identifier
    self.version = version

  def GetMessageFileIdentifier(self):
    """Retrieves the identifier of the associated message file.

    Returns:
      AttributeContainerIdentifier: message file identifier or None when
          not set.
    """
    return self._message_file_identifier

  def SetMessageFileIdentifier(self, message_file_identifier):
    """Sets the identifier of the associated message file.

    Args:
      message_file_identifier (AttributeContainerIdentifier): message file
          identifier.
    """
    self._message_file_identifier = message_file_identifier


manager.AttributeContainersManager.RegisterAttributeContainers([
    EnvironmentVariableArtifact, HostnameArtifact, OperatingSystemArtifact,
    PathArtifact, SourceConfigurationArtifact, SystemConfigurationArtifact,
    TimeZoneArtifact, UserAccountArtifact, WindowsEventLogMessageFileArtifact,
    WindowsEventLogMessageStringArtifact, WindowsEventLogProviderArtifact,
    WindowsMountedDeviceArtifact, WindowsServiceConfigurationArtifact,
    WindowsWevtTemplateEvent])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the bencode parser plugin for uTorrent files."""

import unittest

from plaso.parsers import bencode_parser

from tests.parsers.bencode_plugins import test_lib


class UTorrentPluginTest(test_lib.BencodePluginTestCase):
  """Tests for bencode parser plugin for uTorrent files."""

  def testProcess(self):
    """Tests the Process function."""
    parser = bencode_parser.BencodeParser()
    storage_writer = self._ParseFile(['bencode', 'utorrent'], parser)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 1)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'added_time': '2013-08-03T14:52:12+00:00',
        'caption': 'plaso test',
        'data_type': 'p2p:bittorrent:utorrent',
        'destination': 'e:\\torrent\\files\\plaso test',
        'downloaded_time': '2013-08-03T18:11:35+00:00',
        'modification_times': [
            '2013-08-03T18:11:34+00:00',
            '2013-08-03T16:27:59+00:00'],
        'seedtime': 511}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 0)
    self.CheckEventData(event_data, expected_event_values)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the task-based multi-process processing analysis engine."""

import os
import shutil
import unittest

from plaso.analysis import tagging
from plaso.containers import sessions
from plaso.engine import configurations
from plaso.lib import definitions
from plaso.multi_process import analysis_engine
from plaso.storage import factory as storage_factory

from tests import test_lib as shared_test_lib
from tests.filters import test_lib as filters_test_lib
from tests.multi_process import test_lib


class AnalysisEngineMultiProcessEngineTest(test_lib.MultiProcessingTestCase):
  """Tests for the task-based multi-process processing analysis engine."""

  # pylint: disable=protected-access

  def testInternalAnalyzeEvents(self):
    """Tests the _AnalyzeEvents function."""
    test_file_path = self._GetTestFilePath(['psort_test.plaso'])
    self._SkipIfPathNotExists(test_file_path)

    test_tagging_file_path = self._GetTestFilePath([
        'tagging_file', 'valid.txt'])
    self._SkipIfPathNotExists(test_tagging_file_path)

    session = sessions.Session()

    analysis_plugin = tagging.TaggingAnalysisPlugin()
    analysis_plugin.SetAndLoadTagFile(test_tagging_file_path)

    analysis_plugins = {'tagging': analysis_plugin}

    configuration = configurations.ProcessingConfiguration()
    test_engine = analysis_engine.AnalysisMultiProcessEngine()

    with shared_test_lib.TempDirectory() as temp_directory:
      temp_file = os.path.join(temp_directory, 'storage.plaso')
      shutil.copyfile(test_file_path, temp_file)

      storage_writer = storage_factory.StorageFactory.CreateStorageWriter(
          definitions.DEFAULT_STORAGE_FORMAT)

      test_engine._processing_configuration = configuration
      test_engine._session = session

      test_engine._storage_file_path = temp_directory
      test_engine._StartTaskStorage(definitions.STORAGE_FORMAT_SQLITE)

      test_engine._StartAnalysisProcesses(analysis_plugins)

      storage_writer.Open(path=temp_file)

      try:
        events_counter = test_engine._AnalyzeEvents(
            storage_writer, analysis_plugins)
      finally:
        storage_writer.Close()

      test_engine._StopAnalysisProcesses()

    self.assertIsNotNone(events_counter)
    self.assertEqual(events_counter['Events filtered'], 0)
    self.assertEqual(events_counter['Events processed'], 38)

  # TODO: add test for _CheckStatusAnalysisProcess.
  # TODO: add test for _StartAnalysisProcesses.
  # TODO: add test for _StatusUpdateThreadMain.
  # TODO: add test for _StopAnalysisProcesses.
  # TODO: add test for _UpdateProcessingStatus.

  def testAnalyzeEvents(self):
    """Tests the AnalyzeEvents function."""
    test_file_path = self._GetTestFilePath(['psort_test.plaso'])
    self._SkipIfPathNotExists(test_file_path)

    test_tagging_file_path = self._GetTestFilePath([
        'tagging_file', 'valid.txt'])
    self._SkipIfPathNotExists(test_tagging_file_path)

    session = sessions.Session()

    data_location = ''

    analysis_plugin = tagging.TaggingAnalysisPlugin()
    analysis_plugin.SetAndLoadTagFile(test_tagging_file_path)

    analysis_plugins = {'tagging': analysis_plugin}

    configuration = configurations.ProcessingConfiguration()
    test_engine = analysis_engine.AnalysisMultiProcessEngine()

    with shared_test_lib.TempDirectory() as temp_directory:
      temp_file = os.path.join(temp_directory, 'storage.plaso')
      shutil.copyfile(test_file_path, temp_file)

      storage_writer = storage_factory.StorageFactory.CreateStorageWriter(
          definitions.DEFAULT_STORAGE_FORMAT)

      storage_writer.Open(path=temp_file)

      try:
        number_of_reports = storage_writer.GetNumberOfAttributeContainers(
            'analysis_report')
        self.assertEqual(number_of_reports, 2)

        test_engine.AnalyzeEvents(
            session, storage_writer, data_location, analysis_plugins,
            configuration, storage_file_path=temp_directory)

        number_of_reports = storage_writer.GetNumberOfAttributeContainers(
            'analysis_report')
        self.assertEqual(number_of_reports, 3)

      finally:
        storage_writer.Close()

  def testAnalyzeEventsWithEventFilter(self):
    """Tests the AnalyzeEvents function with an event filter."""
    test_file_path = self._GetTestFilePath(['psort_test.plaso'])
    self._SkipIfPathNotExists(test_file_path)

    test_tagging_file_path = self._GetTestFilePath([
        'tagging_file', 'valid.txt'])
    self._SkipIfPathNotExists(test_tagging_file_path)

    session = sessions.Session()

    data_location = ''

    analysis_plugin = tagging.TaggingAnalysisPlugin()
    analysis_plugin.SetAndLoadTagFile(test_tagging_file_path)

    analysis_plugins = {'tagging': analysis_plugin}

    configuration = configurations.ProcessingConfiguration()
    test_engine = analysis_engine.AnalysisMultiProcessEngine()
    test_filter = filters_test_lib.TestEventFilter()

    with shared_test_lib.TempDirectory() as temp_directory:
      temp_file = os.path.join(temp_directory, 'storage.plaso')
      shutil.copyfile(test_file_path, temp_file)

      storage_writer = storage_factory.StorageFactory.CreateStorageWriter(
          definitions.DEFAULT_STORAGE_FORMAT)

      storage_writer.Open(path=temp_file)

      try:
        number_of_reports = storage_writer.GetNumberOfAttributeContainers(
            'analysis_report')
        self.assertEqual(number_of_reports, 2)

        test_engine.AnalyzeEvents(
            session, storage_writer, data_location, analysis_plugins,
            configuration, event_filter=test_filter,
            storage_file_path=temp_directory)

        number_of_reports = storage_writer.GetNumberOfAttributeContainers(
            'analysis_report')
        self.assertEqual(number_of_reports, 3)

      finally:
        storage_writer.Close()

  # TODO: add bogus data location test.


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests for the Redis storage."""

import unittest

try:
  # pylint: disable=ungrouped-imports
  import fakeredis
  import redis
  from plaso.storage.redis import redis_store
except ModuleNotFoundError:
  redis = None

from plaso.containers import events
from plaso.containers import sessions
from plaso.containers import tasks

from tests.containers import test_lib as containers_test_lib
from tests.storage import test_lib


@unittest.skipIf(redis is None, 'missing redis support')
class RedisAttributeContainerStoreTest(test_lib.StorageTestCase):
  """Tests for the Redis storage object."""

  # pylint: disable=protected-access

  _REDIS_URL = 'redis://127.0.0.1/0'

  def _CreateRedisClient(self):
    """Creates a Redis client for testing.

    This method will attempt to use a Redis server listening on localhost and
    fallback to a fake Redis client if no server is available or the connection
    timed out.

    Returns:
      Redis: a Redis client.
    """
    try:
      redis_client = redis.from_url(self._REDIS_URL, socket_timeout=60)
      redis_client.ping()
    except redis.exceptions.ConnectionError:
      redis_client = fakeredis.FakeStrictRedis()

    return redis_client

  def _RemoveSessionData(self, redis_client, session_identifier):
    """Removes the session data after testing.

    Args:
      redis_client (Redis): an open Redis client.
      session_identifier (str): the identifier of the session the tasks are
          part of.
    """
    redis_hash_pattern = '{0:s}-*'.format(session_identifier)

    for redis_hash_name in redis_client.keys(redis_hash_pattern):
      redis_client.delete(redis_hash_name)

  def testGetRedisHashName(self):
    """Tests the _GetRedisHashName function."""
    redis_client = self._CreateRedisClient()

    session = sessions.Session()
    task = tasks.Task(session_identifier=session.identifier)

    event_data_stream = events.EventDataStream()

    test_store = redis_store.RedisAttributeContainerStore()
    test_store.Open(
        redis_client=redis_client, session_identifier=task.session_identifier,
        task_identifier=task.identifier)

    try:
      redis_hash_name = test_store._GetRedisHashName(
          event_data_stream.CONTAINER_TYPE)

      expected_redis_hash_name = '{0:s}-{1:s}-{2:s}'.format(
          task.session_identifier, task.identifier,
          event_data_stream.CONTAINER_TYPE)
      self.assertEqual(redis_hash_name, expected_redis_hash_name)

    finally:
      test_store.Close()

      self._RemoveSessionData(redis_client, session.identifier)

  # TODO: add tests for _GetFinalizationKey
  # TODO: add tests for _RaiseIfNotReadable
  # TODO: add tests for _RaiseIfNotWritable
  # TODO: add tests for _SetClientName

  def testWriteExistingAttributeContainer(self):
    """Tests the _WriteExistingAttributeContainer function."""
    redis_client = self._CreateRedisClient()

    session = sessions.Session()
    task = tasks.Task(session_identifier=session.identifier)

    test_store = redis_store.RedisAttributeContainerStore()
    test_store.Open(
        redis_client=redis_client, session_identifier=session.identifier,
        task_identifier=task.identifier)

    try:
      event_data_stream = events.EventDataStream()

      number_of_containers = test_store.GetNumberOfAttributeContainers(
          event_data_stream.CONTAINER_TYPE)
      self.assertEqual(number_of_containers, 0)

      test_store._WriteNewAttributeContainer(event_data_stream)

      number_of_containers = test_store.GetNumberOfAttributeContainers(
          event_data_stream.CONTAINER_TYPE)
      self.assertEqual(number_of_containers, 1)

      test_store._WriteExistingAttributeContainer(event_data_stream)

      number_of_containers = test_store.GetNumberOfAttributeContainers(
          event_data_stream.CONTAINER_TYPE)
      self.assertEqual(number_of_containers, 1)

    finally:
      test_store.Close()

      self._RemoveSessionData(redis_client, session.identifier)

  def testWriteNewAttributeContainer(self):
    """Tests the _WriteNewAttributeContainer method."""
    redis_client = self._CreateRedisClient()

    session = sessions.Session()
    task = tasks.Task(session_identifier=session.identifier)

    test_store = redis_store.RedisAttributeContainerStore()
    test_store.Open(
        redis_client=redis_client, session_identifier=task.session_identifier,
        task_identifier=task.identifier)

    try:
      event_data_stream = events.EventDataStream()

      number_of_containers = test_store.GetNumberOfAttributeContainers(
          event_data_stream.CONTAINER_TYPE)
      self.assertEqual(number_of_containers, 0)

      test_store._WriteNewAttributeContainer(event_data_stream)

      number_of_containers = test_store.GetNumberOfAttributeContainers(
          event_data_stream.CONTAINER_TYPE)
      self.assertEqual(number_of_containers, 1)

    finally:
      test_store.Close()

      self._RemoveSessionData(redis_client, session.identifier)

  def testAddAttributeContainer(self):
    """Tests the AddAttributeContainer method."""
    redis_client = self._CreateRedisClient()

    session = sessions.Session()
    task = tasks.Task(session_identifier=session.identifier)

    test_store = redis_store.RedisAttributeContainerStore()
    test_store.Open(
        redis_client=redis_client, session_identifier=task.session_identifier,
        task_identifier=task.identifier)

    try:
      event_data_stream = events.EventDataStream()

      number_of_containers = test_store.GetNumberOfAttributeContainers(
          event_data_stream.CONTAINER_TYPE)
      self.assertEqual(number_of_containers, 0)

      test_store.AddAttributeContainer(event_data_stream)

      number_of_containers = test_store.GetNumberOfAttributeContainers(
          event_data_stream.CONTAINER_TYPE)
      self.assertEqual(number_of_containers, 1)

    finally:
      test_store.Close()

      self._RemoveSessionData(redis_client, session.identifier)

    with self.assertRaises(IOError):
      test_store.AddAttributeContainer(event_data_stream)

  # TODO: add tests for _WriteStorageMetadata

  def testGetAttributeContainerByIdentifier(self):
    """Tests the GetAttributeContainerByIdentifier method."""
    redis_client = self._CreateRedisClient()

    session = sessions.Session()
    task = tasks.Task(session_identifier=session.identifier)

    test_store = redis_store.RedisAttributeContainerStore()
    test_store.Open(
        redis_client=redis_client, session_identifier=task.session_identifier,
        task_identifier=task.identifier)

    try:
      event_data_stream = events.EventDataStream()

      test_store.AddAttributeContainer(event_data_stream)
      identifier = event_data_stream.GetIdentifier()

      container = test_store.GetAttributeContainerByIdentifier(
          event_data_stream.CONTAINER_TYPE, identifier)
      self.assertIsNotNone(container)

      identifier.sequence_number = 99

      container = test_store.GetAttributeContainerByIdentifier(
          event_data_stream.CONTAINER_TYPE, identifier)
      self.assertIsNone(container)

    finally:
      test_store.Close()

      self._RemoveSessionData(redis_client, session.identifier)

  def testGetAttributeContainerByIndex(self):
    """Tests the GetAttributeContainerByIndex function."""
    redis_client = self._CreateRedisClient()

    session = sessions.Session()
    task = tasks.Task(session_identifier=session.identifier)

    test_store = redis_store.RedisAttributeContainerStore()
    test_store.Open(
        redis_client=redis_client, session_identifier=task.session_identifier,
        task_identifier=task.identifier)

    try:
      event_data_stream = events.EventDataStream()

      container = test_store.GetAttributeContainerByIndex(
          event_data_stream.CONTAINER_TYPE, 0)
      self.assertIsNone(container)

      test_store.AddAttributeContainer(event_data_stream)

      container = test_store.GetAttributeContainerByIndex(
          event_data_stream.CONTAINER_TYPE, 0)
      self.assertIsNotNone(container)

    finally:
      test_store.Close()

      self._RemoveSessionData(redis_client, session.identifier)

  def testGetAttributeContainers(self):
    """Tests the GetAttributeContainers method."""
    redis_client = self._CreateRedisClient()

    session = sessions.Session()
    task = tasks.Task(session_identifier=session.identifier)

    test_store = redis_store.RedisAttributeContainerStore()
    test_store.Open(
        redis_client=redis_client, session_identifier=task.session_identifier,
        task_identifier=task.identifier)

    try:
      event_data_stream = events.EventDataStream()
      event_data_stream.md5_hash = '8f0bf95a7959baad9666b21a7feed79d'

      containers = list(test_store.GetAttributeContainers(
          event_data_stream.CONTAINER_TYPE))
      self.assertEqual(len(containers), 0)

      test_store.AddAttributeContainer(event_data_stream)

      containers = list(test_store.GetAttributeContainers(
          event_data_stream.CONTAINER_TYPE))
      self.assertEqual(len(containers), 1)

      filter_expression = 'md5_hash == "8f0bf95a7959baad9666b21a7feed79d"'
      containers = list(test_store.GetAttributeContainers(
          event_data_stream.CONTAINER_TYPE,
          filter_expression=filter_expression))
      self.assertEqual(len(containers), 1)

      filter_expression = 'md5_hash != "8f0bf95a7959baad9666b21a7feed79d"'
      containers = list(test_store.GetAttributeContainers(
         event_data_stream.CONTAINER_TYPE,
         filter_expression=filter_expression))
      self.assertEqual(len(containers), 0)

    finally:
      test_store.Close()

      self._RemoveSessionData(redis_client, session.identifier)

  def testGetNumberOfAttributeContainers(self):
    """Tests the GetNumberOfAttributeContainers function."""
    redis_client = self._CreateRedisClient()

    session = sessions.Session()
    task = tasks.Task(session_identifier=session.identifier)

    test_store = redis_store.RedisAttributeContainerStore()
    test_store.Open(
        redis_client=redis_client, session_identifier=task.session_identifier,
        task_identifier=task.identifier)

    try:
      event_data_stream = events.EventDataStream()

      number_of_containers = test_store.GetNumberOfAttributeContainers(
          event_data_stream.CONTAINER_TYPE)
      self.assertEqual(number_of_containers, 0)

      test_store.AddAttributeContainer(event_data_stream)

      number_of_containers = test_store.GetNumberOfAttributeContainers(
          event_data_stream.CONTAINER_TYPE)
      self.assertEqual(number_of_containers, 1)

    finally:
      test_store.Close()

      self._RemoveSessionData(redis_client, session.identifier)

  def testGetSortedEvents(self):
    """Tests the GetSortedEvents method."""
    redis_client = self._CreateRedisClient()

    session = sessions.Session()
    task = tasks.Task(session_identifier=session.identifier)

    test_store = redis_store.RedisAttributeContainerStore()
    test_store.Open(
        redis_client=redis_client, session_identifier=task.session_identifier,
        task_identifier=task.identifier)

    try:
      for event, _, _ in containers_test_lib.CreateEventsFromValues(
          self._TEST_EVENTS):
        test_store.AddAttributeContainer(event)

      retrieved_events = list(test_store.GetSortedEvents())
      self.assertEqual(len(retrieved_events), 4)

    finally:
      test_store.Close()

      self._RemoveSessionData(redis_client, session.identifier)

  def testHasAttributeContainers(self):
    """Tests the HasAttributeContainers method."""
    redis_client = self._CreateRedisClient()

    session = sessions.Session()
    task = tasks.Task(session_identifier=session.identifier)

    test_store = redis_store.RedisAttributeContainerStore()
    test_store.Open(
        redis_client=redis_client, session_identifier=task.session_identifier,
        task_identifier=task.identifier)

    try:
      event_data_stream = events.EventDataStream()

      result = test_store.HasAttributeContainers(
          event_data_stream.CONTAINER_TYPE)
      self.assertFalse(result)

      test_store.AddAttributeContainer(event_data_stream)

      result = test_store.HasAttributeContainers(
          event_data_stream.CONTAINER_TYPE)
      self.assertTrue(result)

    finally:
      test_store.Close()

      self._RemoveSessionData(redis_client, session.identifier)

  # TODO: add tests for Open and Close

  def testUpdateAttributeContainer(self):
    """Tests the UpdateAttributeContainer function."""
    redis_client = self._CreateRedisClient()

    session = sessions.Session()
    task = tasks.Task(session_identifier=session.identifier)

    test_store = redis_store.RedisAttributeContainerStore()
    test_store.Open(
        redis_client=redis_client, session_identifier=task.session_identifier,
        task_identifier=task.identifier)

    try:
      event_data_stream = events.EventDataStream()

      number_of_containers = test_store.GetNumberOfAttributeContainers(
          event_data_stream.CONTAINER_TYPE)
      self.assertEqual(number_of_containers, 0)

      test_store.AddAttributeContainer(event_data_stream)

      number_of_containers = test_store.GetNumberOfAttributeContainers(
          event_data_stream.CONTAINER_TYPE)
      self.assertEqual(number_of_containers, 1)

      test_store.UpdateAttributeContainer(event_data_stream)

      number_of_containers = test_store.GetNumberOfAttributeContainers(
          event_data_stream.CONTAINER_TYPE)
      self.assertEqual(number_of_containers, 1)

    finally:
      test_store.Close()

      self._RemoveSessionData(redis_client, session.identifier)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*
"""Tests for the JSON-L parser plugin for AWS CloudTrail log files."""

import unittest

from plaso.parsers.jsonl_plugins import aws_cloudtrail_log

from tests.parsers.jsonl_plugins import test_lib


class AWSCloudTrailLogJSONLPluginTest(test_lib.JSONLPluginTestCase):
  """Tests for the JSON-L parser plugin for AWS CloudTrail log files."""

  def testProcess(self):
    """Tests the Process function."""
    plugin = aws_cloudtrail_log.AWSCloudTrailLogJSONLPlugin()
    storage_writer = self._ParseJSONLFileWithPlugin(
        ['aws_cloudtrail.jsonl'], plugin)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 6)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'access_key': '0123456789ABCDEFGHIJ',
        'event_name': 'DescribeInstances',
        'event_source': 'ec2.amazonaws.com',
        'recorded_time': '2022-02-08T09:23:37.000000+11:00',
        'source_ip': '1.2.3.4',
        'user_identity_arn': 'arn:aws:iam::012345678901:user/fakeusername',
        'user_name': 'fakeusername'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 0)
    self.CheckEventData(event_data, expected_event_values)


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for the OXML plugin."""

import unittest

from plaso.parsers.czip_plugins import oxml

from tests.parsers.czip_plugins import test_lib


class OXMLTest(test_lib.CompoundZIPPluginTestCase):
  """Tests for the OXML plugin."""

  # pylint: disable=protected-access

  _PROPERTIES_XML_DATA = ''.join([
      '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>',
      ('<cp:coreProperties xmlns:cp="http://schemas.openxmlformats.org/'
       'package/2006/metadata/core-properties" xmlns:dc="http://purl.org/'
       'dc/elements/1.1/" xmlns:dcterms="http://purl.org/dc/terms/" '
       'xmlns:dcmitype="http://purl.org/dc/dcmitype/" '
       'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">'),
      '<dc:creator>Nides</dc:creator>',
      '<cp:lastModifiedBy>Nides</cp:lastModifiedBy>',
      '<cp:revision>3</cp:revision>',
      ('<dcterms:created xsi:type="dcterms:W3CDTF">2012-11-07T23:29:00.1234567Z'
       '</dcterms:created>'),
      ('<dcterms:modified xsi:type="dcterms:W3CDTF">2013-08-25T22:18:00Z'
       '</dcterms:modified>'),
      '</cp:coreProperties>'])

  _RELATIONSHIPS_XML_DATA = ''.join([
      '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>',
      ('<Relationships xmlns="http://schemas.openxmlformats.org/package/2006/'
       'relationships">'),
      ('<Relationship Id="rId3" Type="http://schemas.openxmlformats.org/'
       'package/2006/relationships/metadata/core-properties" Target="docProps/'
       'core.xml"/>'),
      ('<Relationship Id="rId2" Type="http://schemas.openxmlformats.org/'
       'package/2006/relationships/metadata/thumbnail" Target="docProps/'
       'thumbnail.emf"/>'),
      ('<Relationship Id="rId1" Type="http://schemas.openxmlformats.org/'
       'officeDocument/2006/relationships/officeDocument" Target="word/'
       'document.xml"/>'),
      ('<Relationship Id="rId4" Type="http://schemas.openxmlformats.org/'
       'officeDocument/2006/relationships/extended-properties" '
       'Target="docProps/app.xml"/>'),
      '</Relationships>'])

  # TODO: add tests for _FormatPropertyName.

  def testParsePropertiesXMLFile(self):
    """Tests the _ParsePropertiesXMLFile function."""
    plugin = oxml.OpenXMLPlugin()

    expected_properties = {
        'author': 'Nides',
        'created': '2012-11-07T23:29:00.1234567Z',
        'last_saved_by': 'Nides',
        'modified': '2013-08-25T22:18:00Z',
        'revision_number': '3'}

    properties = plugin._ParsePropertiesXMLFile(self._PROPERTIES_XML_DATA)
    self.assertEqual(properties, expected_properties)

  def testParseRelationshipsXMLFile(self):
    """Tests the _ParseRelationshipsXMLFile function."""
    plugin = oxml.OpenXMLPlugin()

    expected_property_files = ['docProps/core.xml', 'docProps/app.xml']

    property_files = plugin._ParseRelationshipsXMLFile(
        self._RELATIONSHIPS_XML_DATA)
    self.assertEqual(property_files, expected_property_files)

  def testParseFileObject(self):
    """Tests the ParseFileObject function."""
    plugin = oxml.OpenXMLPlugin()
    storage_writer = self._ParseZIPFileWithPlugin(['Document.docx'], plugin)

    number_of_event_data = storage_writer.GetNumberOfAttributeContainers(
        'event_data')
    self.assertEqual(number_of_event_data, 1)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'extraction_warning')
    self.assertEqual(number_of_warnings, 0)

    number_of_warnings = storage_writer.GetNumberOfAttributeContainers(
        'recovery_warning')
    self.assertEqual(number_of_warnings, 0)

    expected_event_values = {
        'application': 'Microsoft Office Word',
        'application_version': '14.0000',
        'author': 'Nides',
        'creation_time': '2012-11-07T23:29:00.000000+00:00',
        'data_type': 'openxml:metadata',
        'edit_duration': 1385,
        'hyperlinks_changed': False,
        'last_printed_time': None,
        'last_saved_by': 'Nides',
        'links_up_to_date': False,
        'modification_time': '2013-08-25T22:18:00.000000+00:00',
        'number_of_characters': 13,
        'number_of_characters_with_spaces': 14,
        'number_of_clips': None,
        'number_of_hidden_slides': None,
        'number_of_lines': 1,
        'number_of_pages': 1,
        'number_of_paragraphs': 1,
        'number_of_slides': None,
        'number_of_words': 2,
        'revision_number': 3,
        'scale': False,
        'security_flags': 0,
        'template': 'Normal.dotm'}

    event_data = storage_writer.GetAttributeContainerByIndex('event_data', 0)
    self.CheckEventData(event_data, expected_event_values)


if __name__ == '__main__':
  unittest.main()
